import os
import pickle
import multiprocessing
import itertools
import shutil
import subprocess
import glob


import pandas as pnd
from Bio import SeqIO, SeqRecord, Seq


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import check_cached
from ..commons import create_summary
from ..commons import update_pam
from ..commons import extract_aa_seq_from_genome
from ..commons import get_blast_header



def update_seq_to_coords(logger): 
    
    
    # load the previously created species_to_proteome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
        
    
    # get the accessions (passing the quality filters):
    accessions = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            accessions.append(accession)
    

    # create an updateed seq_to_coords dict: 
    with open('working/rec_broken/seq_to_coords.pickle', 'rb') as handler:
        seq_to_coords_update = pickle.load(handler)
    logger.debug(f'rec_masking: seq_to_coords: starting from {len(seq_to_coords_update.values())} sequences.')

        
    # now add the new seqs (recovered by this module): 
    for accession in accessions: 
        results_df = pnd.read_csv(f'working/rec_masking/results/{accession}.csv', index_col=0)
        for index, row in results_df.iterrows():
            seq_to_coords_update[row['ID']] = {'accession': row['accession'], 'contig': row['contig'], 'strand': row['strand'], 'start': row['start'], 'end': row['end']}
    logger.debug(f'rec_masking: seq_to_coords: {len(seq_to_coords_update.values())} sequences after the addition of new IDs.')

    
    # save the update dictionary: 
    with open('working/rec_masking/seq_to_coords.pickle', 'wb') as file:
        pickle.dump(seq_to_coords_update, file)



def genome_masking(genome, seq_to_coords):

    
    # get the basename without extension:
    basename = os.path.basename(genome)
    accession, _ = os.path.splitext(basename)

    
    # parse each genome:
    sr_list = []
    with open(genome, 'r') as r_handler:                  
        for seqrecord in SeqIO.parse(r_handler, "fasta"):
            contig = seqrecord.id
            seq = seqrecord.seq
            seq_masked = seqrecord.seq  # to be updated
            
            
            # mask the contig from its genes: 
            for seq in seq_to_coords.keys(): 
                if seq_to_coords[seq]['accession'] == accession:
                    if seq_to_coords[seq]['contig'] == contig: 
                        start, end = seq_to_coords[seq]['start'], seq_to_coords[seq]['end']
                        cds_len = end - start  # in this case, 'end' is always greater then 'start'.
                        gene_masked = Seq.Seq(''.join(['N' for i in range(cds_len)]))
                        seq_masked = seq_masked[:start] + gene_masked + seq_masked[end:]
                    
            
            # save the masked squences: 
            sr = SeqRecord.SeqRecord(seq_masked, id=contig, description='')
            sr_list.append(sr)
    with open(f'working/rec_masking/masked_assemblies/{accession}.masked.fna', 'w') as w_handler:
        count = SeqIO.write(sr_list, w_handler, "fasta")



def task_recmasking(genome, args):
    
    
    # retrive the arguments:
    pam = args['pam']
    cluster_to_rep = args['cluster_to_rep']
    rep_to_aaseq = args['rep_to_aaseq']
    acc_to_suffix = args['acc_to_suffix']
    seq_to_coords = args['seq_to_coords']
    
    
    # get the basename without extension:
    basename = os.path.basename(genome)
    accession, _ = os.path.splitext(basename)
    
    
    # create a query file for each genome: 
    sr_list = []
    with open(f'working/rec_masking/queries/{accession}.query.faa', 'w') as w_handler: 
        for cluster in pam.index:
            cell = pam.loc[cluster, accession]
            if type(cell) == float:  # include only empty clusters
                rep = cluster_to_rep[cluster]
                seq = Seq.Seq(rep_to_aaseq[rep])
                sr = SeqRecord.SeqRecord(seq, id=cluster, description='')
                sr_list.append(sr) 
        count = SeqIO.write(sr_list, w_handler, "fasta")


    # mask the genome from its genes:
    response = genome_masking(genome, seq_to_coords)


    # create a blast database for the genome:
    os.makedirs(f'working/rec_masking/databases/{accession}/', exist_ok=True)
    shutil.copyfile(f'working/rec_masking/masked_assemblies/{accession}.masked.fna', f'working/rec_masking/databases/{accession}/{accession}.masked.fna')  # just the content, not the permissions.
    command = f"""makeblastdb -in working/rec_masking/databases/{accession}/{accession}.masked.fna -dbtype nucl -parse_seqids"""  # '-parse_seqids' is required for 'blastdbcmd'.
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    # perform the blast search:
    command = f'''tblastn \
        -query working/rec_masking/queries/{accession}.query.faa \
        -db working/rec_masking/databases/{accession}/{accession}.masked.fna \
        -out working/rec_masking/alignments/{accession}.tsv \
        -outfmt "6 {get_blast_header()}"
    '''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    # read the alignment: 
    colnames = f'{get_blast_header()}'.split(' ')
    alignment = pnd.read_csv(f'working/rec_masking/alignments/{accession}.tsv', sep='\t', names=colnames )
    alignment['qcov'] = round((alignment['qend'] -  alignment['qstart'] +1)/ alignment['qlen'] * 100, 1)
    alignment['scov'] = round((alignment['send'] -  alignment['sstart'] +1)/ alignment['slen'] * 100, 1)
    
    
    # instantiate key objects: 
    cnt_good_genes = 0
    new_rows = []  # new rows for the sequences_df
    df_result = []  # using the common formatting
    alignment_filtered = []  # future dataframe (blast tbl + seq ID)
    
    
    # group the hsps by cluster:
    cluster_to_indexes = alignment.groupby(['qseqid']).groups
    for cluster in cluster_to_indexes.keys():
        # isolate results for this cluster: 
        indexes = cluster_to_indexes[cluster]
        alignment_cluster = alignment.iloc[indexes, ]
        
        
        # filter using the same thresholds of cd-hit
        alignment_cluster = alignment_cluster[(alignment_cluster['pident'] >= 90) & (alignment_cluster['qcov'] >= 70)]
        
        
        # parse each good hsp: 
        for index, row in alignment_cluster.iterrows():
            cnt_good_genes += 1 
            refound_gid = f"{acc_to_suffix[accession]}_refound_{cnt_good_genes}"


            # retrieve the sequence from the genome:
            start = int(row['sstart'])
            end = int(row['send'])
            contig = row["sseqid"]
            contig = contig.split('|')[1]   #split() because blast but some obscure formatting, eg: gb|NIGV01000003.1| for NIGV01000003.1.
            strand = '+'
            if start > end: # if on the other strand, invert the positions. 
                strand = '-'
                start, end = end, start
            curr_seq_translated, curr_seq_translated_tostop = \
                extract_aa_seq_from_genome(
                    f'working/rec_masking/databases/{accession}/{accession}.masked.fna',
                    contig, strand, start, end) 
            
            
            # if premature stop, sign the ID
            if len(curr_seq_translated_tostop) / len(curr_seq_translated) < 0.95 :
                refound_gid = refound_gid + '_stop'


            # populate the results dataframe:
            df_result.append({
                'ID': refound_gid, 'cluster': cluster, 
                'accession': accession, 'contig': contig, 'strand': strand, 
                'start': start, 'end': end,
            })
            
            
            # populate the alignment_filtered dataframe
            align_filt_row = row.to_dict()
            align_filt_row['ID'] = refound_gid
            alignment_filtered.append(align_filt_row)
            
            
            # populate the sequences dataframe
            new_rows.append({'cds': refound_gid, 'accession': accession, 'aaseq': str(curr_seq_translated)})


    # save the filtered hsps for this genome:
    alignment_filtered = pnd.DataFrame.from_records(alignment_filtered)
    alignment_filtered.to_csv(f'working/rec_masking/alignments_filtered/{accession}.csv')
    
            
    # save results for this genome:
    df_result = pnd.DataFrame.from_records(df_result)
    df_result.to_csv(f'working/rec_masking/results/{accession}.csv')
    
    
    # return new rows for the sequences_df
    return new_rows



def recovery_masking(logger, cores):
    
    
    # some log messages:
    logger.info("Recovering the genes using genome masking...")
    
    
    # create sub-directories without overwriting:
    os.makedirs('working/rec_masking/', exist_ok=True)
    os.makedirs('working/rec_masking/queries/', exist_ok=True)
    os.makedirs('working/rec_masking/masked_assemblies/', exist_ok=True)
    os.makedirs('working/rec_masking/databases/', exist_ok=True)
    os.makedirs(f'working/rec_masking/alignments/', exist_ok=True)
    os.makedirs(f'working/rec_masking/alignments_filtered/', exist_ok=True)
    os.makedirs('working/rec_masking/results/', exist_ok=True)
    
    
    # check if it's everything pre-computed
    response = check_cached(
        logger, pam_path='working/rec_masking/pam.csv',
        summary_path='working/rec_masking/summary.csv',
        imp_files = [
            'working/rec_masking/sequences.csv',
            'working/rec_masking/seq_to_coords.pickle',])
    if response == 0: 
        return 0
    
    
    # load the assets to form the args dictionary:
    pam = pnd.read_csv('working/rec_broken/pam.csv', index_col=0)
    with open('working/clustering/cluster_to_rep.pickle', 'rb') as handler:
        cluster_to_rep = pickle.load(handler)
    with open('working/clustering/rep_to_aaseq.pickle', 'rb') as handler:
        rep_to_aaseq = pickle.load(handler)
    with open('working/clustering/acc_to_suffix.pickle', 'rb') as handler:
        acc_to_suffix = pickle.load(handler)
    with open('working/rec_broken/seq_to_coords.pickle', 'rb') as handler:
        seq_to_coords = pickle.load(handler)


    # load the previously created species_to_genome: 
    with open('working/genomes/species_to_genome.pickle', 'rb') as handler:
        species_to_genome = pickle.load(handler)


    # create items for parallelization: 
    items = []
    for species in species_to_genome.keys(): 
        for genome in species_to_genome[species]: 
            items.append(genome)
    

    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['cds', 'accession', 'aaseq']), 
            itertools.repeat('cds'), 
            itertools.repeat(logger), 
            itertools.repeat(task_recmasking),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'pam': pam, 'cluster_to_rep': cluster_to_rep, 'rep_to_aaseq': rep_to_aaseq, 'acc_to_suffix': acc_to_suffix, 'seq_to_coords': seq_to_coords}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # save tabular results:
    sequences_df = pnd.read_csv('working/rec_broken/sequences.csv', index_col=0)
    sequences_df_updated = pnd.concat([sequences_df, all_df_combined], axis=0)
    sequences_df_updated.to_csv('working/rec_masking/sequences.csv')
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # update the pam:
    update_pam(logger, module_dir='working/rec_masking', pam=pam)
    
    
    # update the seq to coordinates dictionary
    update_seq_to_coords(logger)
    
    
    # create the summary:
    create_summary(logger, module_dir='working/rec_masking')
    
    
    return 0
    
    


"""
Please note that all the functions from ``gempipe.curate.*`` are internally called also from ``gempipe``. 
Therefore, for example, to use the `perform_gapfilling()` function from ``gempipe.curate.gaps``, it's sufficient to write::

    import gempipe
    
    gempipe.perform_gapfilling(model, universe)
"""



import os
import glob
from importlib import resources


import pandas as pnd
import cobra


from ..commons import read_refmodel
from ..commons import get_media_definitions
from ..commons import apply_json_medium
from ..commons import check_panmodel_growth
from ..commons import strenghten_uptakes


from gempipe.recon.networkrec import filter_alignment
from gempipe.recon.networkrec import get_gene_scores_table
from gempipe.recon.networkrec import get_protein_scores_table
from gempipe.recon.networkrec import get_reaction_scores_table
from gempipe.recon.networkrec import normalize_reaction_scores


from gempipe.recon.refexpansion import add_new_reaction
from gempipe.recon.refexpansion import mancor_to_dict


from gempipe.curate.gaps import perform_gapfilling
from gempipe.curate.gaps import import_from_universe
from gempipe.curate.gaps import check_reactants
from gempipe.curate.gaps import get_objectives
from gempipe.curate.gaps import add_demand



def get_Rset_dataframe(panmodel, reaction_scores, refmodel, universe):
    # The Rset_dataframe is a table listing all the reaction that are currently NOT part of the draft panmodel.
    # These missing reactions may come (1) from the reference model, or (2) from the universe model.
    # The Rset_dataframe will also report lb, ub, n_score (normalized score), and penalty for each of these reactions. 
    
    
    # save the reaction table to the gapfilling folder
    reaction_scores['reaction'] = reaction_scores['reaction'].str[2:]  # remove the 'R_' prefix (eg R_11M3ODO -> 11M3ODO)
    reaction_scores = reaction_scores.set_index('reaction', drop=True, verify_integrity=True)  
    
    
    Rset = set()  # expanding set of processed reactions
    Rset_dataframe = []   # list of dicts, future df. 
    modeled_rids = [r.id for r in panmodel.reactions]
    tabled_rids = list(reaction_scores.index) 
        
    
    # parse the reference and universe model:
    # note: reference reactions have priority over universe reaction as reference comes as first!
    for i, source in enumerate([refmodel, universe]):
        for r in source.reactions:
            if r.id not in modeled_rids:
                if r.id in Rset: 
                    continue  # already coming from the reference
                    
                    
                # some flags: in reference / in reaction scores table:
                in_ref = True if i==0 else False
                in_tab = r.id in tabled_rids
                
                
                # define the nature of this reaction
                is_enzymatic = True
                if len(r.metabolites)==1:  # exclude EX_change , sink, demand reactions. 
                    is_enzymatic = False  
                
                
                # define the (normalized) reaction score for this missing reaction
                n_score = 0  # like 'spontaneous' reactions
                if in_tab: n_score = reaction_scores.loc[r.id, 'normalized_score']
                
                
                # define the penalty for this reaction
                penalty = 1 / (1 + n_score)
                
                
                # populate the dataframe: 
                Rset_dataframe.append({
                    'rid': r.id, 'in_ref': in_ref, 'in_tab': in_tab, 'is_enzymatic': is_enzymatic,
                    'lb': r.lower_bound, 'ub': r.upper_bound, 'n_score': n_score, 'penalty': penalty})
                Rset.add(r.id)
        
            
    # build and save the Rset_dataframe:
    Rset_dataframe = pnd.DataFrame.from_records(Rset_dataframe)
    Rset_dataframe = Rset_dataframe.set_index('rid', drop=True, verify_integrity=True)
    Rset_dataframe = Rset_dataframe.sort_values(by='penalty', ascending=True)  # top scoring on top.
    Rset_dataframe.to_csv('working/gapfilling/Rset_dataframe.csv')
    
    
    return Rset_dataframe



def create_extended_universe(logger, panmodel, Rset_dataframe, refmodel, universe, mancor_filepath):
    # create an 'all_model', starting from the draft pan model, 
    # adding all the missing reactions both from the reference and the universe.
    expanded_universe = panmodel.copy()
    
    # remove all residual genes:
    cobra.manipulation.delete.remove_genes(expanded_universe, [g for g in expanded_universe.genes], remove_reactions=False)
    
    
    # below we use the add_new_reaction() function from refexpansion.py.
    # Therefore we need the same 'addedms_logger', and 'mancor'.
    addedms_logger = open('working/gapfilling/candidate_metabolites.txt', 'w')
    if mancor_filepath != '-': # existence of the file was verified in gempipe.recon
        mancor = mancor_to_dict(logger, mancor_filepath)  # mancor formatting was already veryfied in gempipe.recon
    else: mancor = {'formulas': {}, 'charges': {}, 'reactions': {}, 'blacklist': []}  # emtpy, easier to handle


    # iterate the 'Rset_dataframe' and pick up reactions from the appropriate source:
    reactions_to_add = len(Rset_dataframe)
    cnt_added = 0
    current_percentage = 0
    for rid, row in Rset_dataframe.iterrows(): 
        if row['in_ref'] == True:   
            # then pick up from the reference
            r = refmodel.reactions.get_by_id(rid)
        else: # then pick up from the universe
            r = universe.reactions.get_by_id(rid)
        
        
        # assure that exchange reactions have bounds (0, 1000)
        if len(r.metabolites)==1 and list(r.metabolites)[0].id.endswith('_e'):
            bounds = (0, 1000)
        else: bounds = r.bounds
        
        
        # add the reaction:
        add_new_reaction(r, expanded_universe, addedms_logger, mancor, empty_gpr=True, bounds=bounds)
        cnt_added += 1
            
        
        # log progress
        progress = int(cnt_added/reactions_to_add * 100)
        if progress%10==0 and progress > current_percentage:
            current_percentage = progress
            universe_content = f"G:{len(expanded_universe.genes)}, R:{len(expanded_universe.reactions)}, M:{len(expanded_universe.metabolites)}"
            logger.debug(f"{current_percentage}% ({universe_content})")
        

    # save 'expanded_universe' to feed the later gapfilling.    
    cobra.io.save_json_model(expanded_universe, 'working/gapfilling/expanded_universe.json')
    addedms_logger.close()   # close filestream.
    
    
    return expanded_universe


    
def build_universe_candidates(logger, panmodel, refmodel, refproteome, staining, mancor_filepath):
    # here we basically performe another reaction score computation, like in the 
    # reference-free reconstruction, but this time we do not filter the alignment
    # (or we use really relaxed thresholds):
    logger.debug('Computing new reaction scores used relaxed alignment...')
    alignment_filtered = filter_alignment(logger, identity=10, coverage=40)
    
    
    # load gprm table (gene-to-protein_complex-to-reaction-to-model)
    gprm_table = pnd.read_csv('working/free/gprm_table.csv')
    
    
    # get the 'gene_scores' table:
    gene_scores = get_gene_scores_table(logger, alignment_filtered, gprm_table)
    gene_scores.to_csv('working/gapfilling/gene_scores.csv')
                 
    # get the 'protein_scores' table: 
    protein_scores = get_protein_scores_table(logger, gene_scores)
    protein_scores.to_csv('working/gapfilling/protein_scores.csv')
                 
    # get the 'reaction_scores' table: 
    reaction_scores = get_reaction_scores_table(logger, protein_scores)
    reaction_scores.to_csv('working/gapfilling/reaction_scores.csv')
                 
    # normalize reaction scores:
    reaction_scores_normalized = normalize_reaction_scores(reaction_scores)
    reaction_scores_normalized.to_csv('working/gapfilling/reaction_scores_normalized.csv')
    
    
    # load reference model (if any):
    if refmodel != '-' and refproteome != '-':  # file existance already checked.
        refmodel = read_refmodel(refmodel)
        if type(refmodel)==int: return 1  # an error was raised
    else: refmodel = cobra.Model('empty')
    
    
    # load the universe  (basically it's a wrapper of the recon function):
    from gempipe.recon.networkrec import get_universe_template
    universe = get_universe_template(logger=None, staining=staining)
    
        
    # get the set of universal + reference reactions that are not yet in the draft panmodel
    logger.debug('Gathering candidate gap-filling reactions...')
    Rset_dataframe = get_Rset_dataframe(panmodel, reaction_scores_normalized, refmodel, universe)
        
        
    # add the missing reactions to the panmodel, in order to produce a single, expanded universe.
    logger.debug('Building the expanded universe...')
    expanded_universe = create_extended_universe(logger, panmodel, Rset_dataframe, refmodel, universe, mancor_filepath)
    
    
    return Rset_dataframe, expanded_universe
    
        
    
def check_expuni_growth(logger, expanded_universe, media): 
    
    
    # log some message:
    logger.debug("Testing the growth of the 'expanded_universe' on the provided media...")
    
    
    # iterate the provided media
    for medium_name, medium in media.items():


        # apply the medium recipe:
        response = apply_json_medium(expanded_universe, medium)
        if type(response)==str: 
            logger.error(f"The exchange reaction '{response}' contained in the medium definition '{medium_name}' does not exist in the expanded_universe.")
            return 1


        # verify the growth of the expanded_universe:
        res = expanded_universe.optimize()
        obj_value = res.objective_value
        status = res.status
        can_growth = res.status=='optimal' and obj_value > 0
        logger.debug(f"'expanded_universe' growth on {medium_name}: {can_growth} ({status}, {obj_value}).")


        # raise error if it cannot grow:
        if not can_growth:
            cobra.io.save_json_model(expanded_universe, 'working/gapfilling/expuni_nogrowth.json')
            logger.error(  # log the error message: 
                f"The medium definition '{medium_name}' is unable to support growth of the 'expanded_universe'. " +
                f"A copy of the 'expanded_universe', trying to grow on the '{medium_name}', is saved in 'working/gapfilling/expuni_nogrowth.json' for you to fix the '{medium_name}' definition.")
            return 1
        
        
    return 0
    
    
    
def prio_gapfilling(logger, panmodel, expanded_universe, media, Rset_dataframe):
    
    
    # log some message:
    logger.debug("Performing prioritized gap-filling (using penalties)...")
    rids_to_add = []
    
    
    # create the penalties dictionary
    logger.debug("Creating the penalty dictionary...")
    penalties = {}
    for r in expanded_universe.reactions: 
        try: penalty = Rset_dataframe.loc[r.id, 'penalty']
        # this reaction was not part of the candidate gapfilling set:
        except: penalty = 0  # no penalty for reactions that are already in the draft pan-model.
        penalties[r.id] = penalty
            
            
    # iterate the provided media
    for medium_name, medium in media.items():
        
        
        # apply medium definition to both the models:
        # 'response' cannot be 1, has it was already checked previously in 'check_expuni_growth' for the 'expanded_universe'.
        _ = apply_json_medium(panmodel, medium)
        _ = apply_json_medium(expanded_universe, medium)
        
        
        # get candidate gapfilling reactions: 
        minflux = 0.1* expanded_universe.slim_optimize()
        logger.debug(f"Now gapfilling for medium {medium_name}...")
        timeout = 120
        logger.debug(f"1st attempt: whole gap-filling, timout={timeout}...")
        first_sol_rids = perform_gapfilling(panmodel, expanded_universe, minflux=minflux, nsol=1, penalties=penalties, verbose=True, logger=logger, timeout=timeout)
        
        
        if first_sol_rids == None:  # if errors or timeouts: 
            # try to solve the 'lower integer thresholds' errors etc.
            logger.debug(f"2nd attempt: whole gap-filling, strenghten uptakes, timout={timeout}...")
            
            
            # starting the strenghten_uptakes trick...
            # nested 'with' statement (here + gapfilling) doesn't work, so we create a dictionary to later restore edited bounds:
            exr_ori_pan = strenghten_uptakes(panmodel)
            exr_ori_expuni = strenghten_uptakes(expanded_universe)
            minflux = 0.1* expanded_universe.slim_optimize()
            first_sol_rids = perform_gapfilling(panmodel, expanded_universe, minflux=minflux, nsol=1, penalties=penalties, verbose=True, logger=logger, timeout=timeout)
            
            
            if first_sol_rids == None:  # if (again) errors or timeouts:
                obj_ids =  get_objectives(panmodel)
                logger.debug(f"3rd attempt: splitted gap-filling, strenghten uptakes, timout=None...")
                
                
                all_first_sol_rids = []   # considering all precursors
                for mid in check_reactants(panmodel, rid=obj_ids[0], verbose=False):
                    logger.debug(f"Gap-filling for precursor {mid}...")
                    with expanded_universe:
                        expanded_universe.objective = add_demand(expanded_universe, mid)
                        minflux = 0.1* expanded_universe.slim_optimize()
                    first_sol_rids = perform_gapfilling(
                        panmodel, expanded_universe, minflux=minflux, nsol=1, penalties=penalties, 
                        verbose=True, logger=logger, timeout=None, mid=mid)
                    
                    
                    if first_sol_rids == None:  # if (again) errors or timeouts: 
                        logger.error(f"All gap-filling approaches failed.")
                        return 1
                    else:  # if this precursor was gap-filled: 
                        all_first_sol_rids = all_first_sol_rids  + first_sol_rids
                first_sol_rids = list(set(all_first_sol_rids))  # remove duplicates
                        
                    
            # now restore the medium changes!
            for rid in exr_ori_pan.keys(): panmodel.reactions.get_by_id(rid).lower_bound = exr_ori_pan[rid]
            for rid in exr_ori_expuni.keys(): expanded_universe.reactions.get_by_id(rid).lower_bound = exr_ori_expuni[rid]

        
        # append candidate reactions to the 'rids_to_add' list:
        if first_sol_rids==[]:
            logger.debug(f"Medium {medium_name}: growth was already enabled, no gap-filling reactions needed.")
        else: 
            logger.debug(f"Medium {medium_name}: found these gap-filling reactions: {first_sol_rids}.")
            rids_to_add = rids_to_add + first_sol_rids   # 'rids_to_add' is considering all media
            
            
    # insert all the new reactions:
    rids_to_add = list(set(rids_to_add))
    for rid in rids_to_add:
        import_from_universe(panmodel, expanded_universe, rid, gpr=None)
    
     
    return 0
            
    
    
def check_modeled_ingredients(logger, panmodel, staining, media):
    
    
    # load the universe  (basically it's a wrapper of the recon function):
    from gempipe.recon.networkrec import get_universe_template
    universe = get_universe_template(logger=None, staining=staining)
    
    
    # get the available rids: 
    rids_panmodel = [r.id for r in panmodel.reactions]
    rids_universe = [r.id for r in universe.reactions]
    
    
    # iterate the provided media
    for medium_name, medium in media.items():
        
        for rid in medium.keys():
            if rid not in rids_panmodel: 
                if rid in rids_universe:  # import the EXR if it's available in uni: 
                    import_from_universe(panmodel, universe, rid, gpr=None)
                    logger.debug(f"Transferring '{rid}' from universe to draft panmodel, as it appears in {medium_name} recipe.")
                else:  # EXR not present in uni, raise an error: 
                    logger.error(f"The exchange reaction '{rid}' contained in the medium definition '{medium_name}'" + \
                                 "is not present in the draft panmodel, nor in the universe")
                    return 1
    
    return 0
    
    

def prio_gapfiller(logger, refmodel, refproteome, staining, mancor_filepath, media_filepath, minpanflux):
    
    
    # create subdirs without overwriting
    os.makedirs('working/gapfilling/', exist_ok=True)
    
    
    # some log messages
    logger.info('Now starting the prioritized gap-filling for the draft pan-model...')
    
    
    # load draft panmodel: 
    panmodel = cobra.io.load_json_model('working/duplicates/draft_panmodel.json')
    logger.debug(f"Starting with content: G {len(panmodel.genes)} R {len(panmodel.reactions)} M {len(panmodel.metabolites)}")
    
    
    # get the list of media on which to gap-fill: 
    media = get_media_definitions(logger, media_filepath)
    if type(media)==int: return 1   # we encountered an error.


    # the provided media definitions may include substrates (exr) not yet included in the 'panmodel'.
    # therefore we check that all exr in the recipe are included in the model. 
    # if some exr is missing, we try to import it from the universe. 
    response = check_modeled_ingredients(logger, panmodel, staining, media)
    if response == 1: return 1

    
    # check if the draft panmodel is already able to grow:
    response = check_panmodel_growth(logger, panmodel, media, minpanflux)
    if type(response)==int: return 1   # apply_json_medium() failed.
    if response == True: 
        # draft panmodel already produces biomass on all the provided media, 
        # so there's no need to proceed witht the gap-filling.
        # "working/duplicates/draft_panmodel.json" was already produced by gempipe.recon.draft_reconstruction()
        # and it won't be replaced. 
        logger.debug("No need to perform gap-filling on draft pan-model!")
        return 0
    
    
    # build the condidate gap-filler list (Rset) and the expanded universe
    Rset_dataframe, expanded_universe = build_universe_candidates(logger, panmodel, refmodel, refproteome, staining, mancor_filepath)
    
    
    # check if the 'expanded_universe' can grow in the given list of media.
    response = check_expuni_growth(logger, expanded_universe, media)
    if response == 1: return 1
            
    
    # perform the prioritized gap-filling.
    response = prio_gapfilling(logger, panmodel, expanded_universe, media, Rset_dataframe)
    if response == 1: return 1
    
    
    # OVERWRITE draft pan-model (originally created by gempipe.recon.draft_reconstruction())
    cobra.io.save_json_model(panmodel, f'working/duplicates/draft_panmodel.json')
    logger.debug(f"Ending with content: G {len(panmodel.genes)} R {len(panmodel.reactions)} M {len(panmodel.metabolites)}")
    
    
    return 0
    


import multiprocessing
import itertools
import os
import shutil


import pandas as pnd
import cobra


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results



def task_deriverpam(accession, args):
    
    
    # get the arguments
    outdir = args['outdir']
    skipgf = args['skipgf']
    
    
    # load the model
    if not skipgf:
        ss_model = cobra.io.load_json_model(outdir + f'strain_models_gf/{accession}.json')
    else:  # user requested to skip the strain-specific gap-filling
        ss_model = cobra.io.load_json_model(outdir + f'strain_models/{accession}.json')
    
    
    # extract the reactions from the model
    row_dict = {'accession': accession}
    for r in ss_model.reactions:
        row_dict[r.id] = 1

    
    # return the new dataframe row
    return [row_dict]



def derive_rpam(logger, outdir, cores, panmodel, skipgf):

    
    # log some messages
    logger.info("Creating reaction presence/absence matrix...")
    

    # create items for parallelization:
    strains_table = pnd.read_csv(outdir + 'derive_strains.csv', index_col=0)
    items = strains_table.index.to_list()  # get all the accessions
        
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession'] + [r.id for r in panmodel.reactions]), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_deriverpam),  
            itertools.repeat({'outdir': outdir, 'skipgf': skipgf}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # save the rpam
    rpam = all_df_combined
    # now replace missing values with 0. 
    # In 'older' pandas versions, fillna() was doing silent dowcasting (see https://medium.com/@felipecaballero/deciphering-the-cryptic-futurewarning-for-fillna-in-pandas-2-01deb4e411a1)
    # In 'future' pandas versions, silent downcasting will be forbidden. 
    # In 'middle' pandas versions, there is an option to simulate the future behaviour.
    # What follows should guarantee the old behaviour in 'older' and 'middle' pandas versions:
    try:
        with pnd.option_context('future.no_silent_downcasting', True):
            rpam = rpam.fillna(0).infer_objects()   # infer_objects() will do downcasting like in 'older' pandas versions
    except: # OptionError: "No such keys(s): 'future.no_silent_downcasting'" ==> 'older' pandas version
        rpam = rpam.fillna(0)  
    rpam = rpam.astype(int)  # force from float to int.
    rpam = rpam.T  # transpose: reactions as rows.
    rpam.to_csv(outdir + 'rpam.csv')
    
    
    return 0



def task_derivespecies(spp, args):
    
    
    # get the arguments
    strains_table = args['strains_table']
    panmodel = args['panmodel']
    outdir = args['outdir']
    rpam = args['rpam']
    sbml = args['sbml']
    
    
    # filter accessions for this spp
    species_table = strains_table[strains_table['species']==spp]
    
    
    # filter rpam for this spp
    species_rpam = rpam[species_table.index.to_list()]
    
    
    # get rids not present in all the strains of this species:
    to_delete = species_rpam.index[species_rpam.sum(axis=1) != len(species_table)].to_list()
    
    
    # convert rid to panmodel's reaction
    spp_model = panmodel.copy()
    to_delete = [spp_model.reactions.get_by_id(rid) for rid in to_delete]
    
    
    # delete marked reactions
    spp_model.remove_reactions(to_delete)
    
    
    # get some metrics: 
    n_G = len(spp_model.genes)
    n_R = len(spp_model.reactions)
    n_M = len(spp_model.metabolites)
    
    
    # try the FBA: 
    res = spp_model.optimize()
    obj_value = res.objective_value
    status = res.status
    
    
    # save species specific model to disk
    spp_filename = spp.replace(' ', '_')  # remove spaces for the filename
    cobra.io.save_json_model(spp_model, f'{outdir}/{spp_filename}.json')
    if sbml: cobra.io.write_sbml_model(spp_model, f'{outdir}/{spp_filename}.xml')

    
    # compose the new row
    return [{'species': spp, 'G': n_G, 'R': n_R, 'M': n_M, 'obj_value': obj_value, 'status': status }]



def derive_species_specific(logger, outdir, cores, panmodel, sbml):


    # log some messages
    logger.info("Deriving species-specific models...")
    
   
    # create output dir
    if os.path.exists(outdir + 'species_model/'):
        # always overwriting if already existing
        shutil.rmtree(outdir + 'species_model/')  
    os.makedirs(outdir + 'species_model/', exist_ok=True)
    

    # load the reaction presence/absence table:
    rpam = pnd.read_csv(outdir + 'rpam.csv', index_col=0)
    
    
    # create items for parallelization:
    strains_table = pnd.read_csv(outdir + 'derive_strains.csv', index_col=0)
    items = list(set(strains_table['species'].to_list()))  # get species list
        
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['species', 'G', 'R', 'M', 'obj_value', 'status']), 
            itertools.repeat('species'), 
            itertools.repeat(logger), 
            itertools.repeat(task_derivespecies),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'strains_table': strains_table, 'panmodel': panmodel, 'outdir': outdir + 'species_model/', 'rpam': rpam, 'sbml': sbml}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    # save tabular output:
    all_df_combined.to_csv(outdir + 'derive_species.csv')
    
    
    return 0






import multiprocessing
import time
import os
from importlib import resources


from IPython.display import display
import pandas as pnd
import cobra
from cobra.flux_analysis.gapfilling import GapFiller
from cobra.util.solver import linear_reaction_coefficients



__GAPSCACHE__ = None



def initialize(outdir):
    """Initialize the gempipe.curate API with the outputs coming from ``gempipe recon``.
    
    This function locates the draft pan-GSMM, PAM, and functional annotation table inside the `gempipe recon`` output folder (``-o``/``--outdir``).
    
    Args:
        outdir (str): path to the main output folder of ``gempipe recon`` (``-o``/``--outdir``). 
    
    Returns:
        cobra.Model: draft pan-GSMM to start the manual curation. 
    """
    
    
    # check the existance of the needed files: 
    if outdir.endswith('/')==False:
        outdir = outdir + '/'
    if not os.path.exists(outdir):
        print(f"ERROR: the specified path doesn't exists ({outdir}).")
        return
    panmodel_path = outdir + 'draft_panmodel.json'
    if not os.path.exists(panmodel_path):
        print(f"ERROR: cannot find a draft panmodel at the specified path ({panmodel_path}).")
        return
    pam_path = outdir + 'pam.csv'
    if not os.path.exists(pam_path):
        print(f"ERROR: cannot find a PAM at the specified path ({pam_path}).")
        return
    annot_path = outdir + 'annotation.csv'
    if not os.path.exists(annot_path):
        print(f"ERROR: cannot find a functional annotation table at the specified path ({annot_path}).")
        return
    report_path = outdir + 'report.csv'
    if not os.path.exists(report_path):
        print(f"ERROR: cannot find a report table at the specified path ({report_path}).")
        return
    
    
    # initilize or re-initialize the cache:
    global __GAPSCACHE__
    __GAPSCACHE__ = {}
    print(f"Loading PAM ({pam_path})...")
    __GAPSCACHE__['pam'] = pnd.read_csv(pam_path, index_col=0)
    print(f"Loading functional annotation table ({annot_path})...")
    __GAPSCACHE__['annot'] = pnd.read_csv(annot_path, index_col=0)
    print(f"Loading report table ({report_path})...")
    __GAPSCACHE__['report'] = pnd.read_csv(report_path, index_col=0)
    
    
    print(f"Loading draft pan-GSMM ({panmodel_path})...")
    return cobra.io.load_json_model(panmodel_path)



def get_objectives(model):
    """Get the IDs of the current objective reactions. 
    
    Args:
        model (cobra.Model): target model.
        
    Returns:
        list: IDs of the reactions set as objective.

    """
    
    objs = list(linear_reaction_coefficients(model).keys())
    obj_ids = [obj.id for obj in objs]
    return obj_ids

        
        
def get_solver(model):
    """Get the ID of the solver associated to the model.
    
    Args:
        model (cobra.Model): target model.
        
    Returns:
        str: ID of the solver (for example: ``glpk_exact``).

    """
    
    solver = str(type(model.solver))
    solver = solver.replace('<class ', '')
    solver = solver.replace("'optlang.", '')
    solver = solver.replace("_interface.Model'>", '')
    
    return solver



def remove_rids(model, rids, inverse=False):
    """Remove reactions from the model given a list of reaction IDs.
    
    Args:
        model (cobra.Model): target model.
        rids (list): reaction IDs.
        inverse (bool): if ``True``, reactions IDs contained in `rids` will be the ones to keep and not to remove.

    """
    
    to_delete = []
    for r in model.reactions: 
        if not inverse:
            if r.id in rids:
                to_delete.append(r)
        else:
            if r.id not in rids:
                to_delete.append(r)
    model.remove_reactions(to_delete)


        
def perform_gapfilling(model, universe, mid=None, slim=None, minflux=1.0, exr=False, nsol=3, penalties=None, verbose=True, timeout=None, logger=None, boost=False): 
    """Propose gap-filling solutions for the specified objective. 
    
    It's possible to gap-fill also for the biosynthesis of a specific metabolite.
    
    Args:
        model (cobra.Model): target model to gap-fill.
        universe (cobra.Model): model from which to take new reactions.
        mid (str): gap-fill for the biosynthesis of a specific metabolite having ID `mid`. Will be ignored if ``None``.
        slim (str): try to reduce the complexity of the universe, considering only its reactions carrying non-0 flux. Can be ``FBA`` or ``FVA``. Will be ignored if ``None``.
        minflux (float): minimal flux to grant through the objective reaction.
        nsol (int): number of alternative solutions. 
        exr (bool): whether to allow the opening of new EX_change reactions.
        penalties (dict): dictionary keyed by reaction ID, containing reaction-specific penalties to apply during gap-filling.
        verbose (bool): if False, just return the lisr of reaction IDs without printing any further information.
        timeout (int): max seconds to wait for the gapfilling step. If ``None``, gap-filling won't be temporized. 
        logger (logging.Logger): write exception on a logger instad of using print().
        boost (bool): if ``True``, consider the current nutritive sources as unlimited (for example, if LB of ``EX_fe3_e`` is -0.0075, it will be raised to -1000).
        
    Returns:
        list: IDs of reactions proposed during the 1st solution.
    """
    
    
    def temporized_task(model, universe, minflux, exr, nsol, penalties):

        # launch cobrapy gapfilling: 
        gapfiller = GapFiller(model, universe, 
            lower_bound = minflux,
            demand_reactions = False, 
            exchange_reactions = exr, 
            integer_threshold = 1e-10,  # default: 1e-6.
            penalties = penalties,
        )
        try: solutions = gapfiller.fill(iterations=nsol)
        except Exception as e: 
            return e
        
        return solutions
            
            
    # temporary changes (objective, solver, demands edits are temporary)
    with model, universe: 

        """
        # set new solver if needed (cannot be 'glpk_exact').
        if get_solver(model) != solver: model.solver = solver
        if get_solver(universe) != solver: universe.solver = solver
        """

        # if focusing on a particular biosynthesis
        if mid != None:
            model.objective = add_demand(model, mid)
            universe.objective = add_demand(universe, mid)
            
            
        # if requested, boost the actual sources:
        for r in model.reactions: 
            if len(r.metabolites)==1 and list(r.metabolites)[0].id.endswith('_e'): # if exchange reaction
                if r.lower_bound < 0: 
                    r.lower_bound = -1000


        # if requested, try to reduce the universe complexity:
        if slim == 'FBA':
            fluxes = universe.optimize().fluxes
            rids_to_keep = fluxes[fluxes != 0].index
            remove_rids(universe, rids_to_keep, inverse=True)
        elif slim == 'FVA':
            fluxes = cobra.flux_analysis.flux_variability_analysis(universe, fraction_of_optimum=0.01, loopless=False)
            rids_to_keep = fluxes[(fluxes['minimum']!=0) | (fluxes['maximum']!=0)].index
            remove_rids(universe, rids_to_keep, inverse=True)
            
            
        # "AssertionError: daemonic processes are not allowed to have children"
        # could be raise as child processes created by multiprocessing.Pool are demoniac by deafault,
        # and demoniac processes cannot create their own child processes like the temporized gap-filler. 
        if timeout != None: 
            
            
            # run the gapfilling algorithm inside a timer: 
            # first define the worker for the separate process
            def worker(results_channel):
                result = temporized_task(model, universe, minflux, exr, nsol, penalties)
                results_channel.put(result)


            # then start the separate process
            results_channel = multiprocessing.Queue()
            process = multiprocessing.Process(target=worker, args=(results_channel,))
            process.start()  # start the process
            process.join(timeout) # wait for the process to finish or timeout
        
        
            # if still running after the timeout:
            if process.is_alive():  
                process.terminate()
                e = "Gap-filling timed out, consider to increase the 'timeout' attribute."
                if verbose:  # avoid the error stack trace
                    print("ERROR: cobrapy:", e) if logger==None else logger.debug(f'cobrapy: {e}')  
                    #raise TimeoutError(result)  # show the error trace
                return None
        
        
            else: # the worker finished: retrieve the result or exception from the queue:
                result = results_channel.get()
                if isinstance(result, Exception):  
                    if verbose: 
                        print("ERROR: cobrapy:", result) if logger==None else logger.debug(f'cobrapy: {result}')  
                        #raise Exception(result)  # show the error trace
                    return None  # exit the function
                else:  # a true result
                    solutions = result
                    
                    
        else:  # do not temporize the gap-filling:
            result = temporized_task(model, universe, minflux, exr, nsol, penalties)
            if isinstance(result, Exception):  
                if verbose: 
                    print("ERROR: cobrapy:", result) if logger==None else logger.debug(f'cobrapy: {result}')
                return None
            else:  # a true result
                solutions = result
                

        # iterate the solutions:
        first_sol_rids = []  # rids proposed during the 1st solution
        for i, solution in enumerate(solutions):
            if verbose and logger==None: print(f'Solution {i+1}. Reactions to add: {len(solution)}.')


            # iterate the reactions: 
            counter = 0
            for r in solution: 
                counter += 1
                # Note: this 'r' is not linked to any model. 
                # Indeed, calling r.model, None will be returned. 
                if verbose and logger==None: print(f'{counter} {r.id} {r.name}')

                # populate results with IDs from first solution:
                if i == 0: first_sol_rids.append(r.id)

            # separate solutions with a new line:
            if i+1 != len(solutions): 
                if verbose and logger==None: print()


        return first_sol_rids
        
    

def get_universe(staining='neg'):
    """Return a CarveMe universe. 
    
    Args:
        staining (str): 'pos' or 'neg'.
        
    Returns: 
        cobra.Model: the selected universe.
    """
    
    # basically it's a wrapper of the recon function
    from gempipe.recon.networkrec import get_universe_template
    universe = get_universe_template(logger=None, staining=staining)
    
    return universe



def get_biolog_mappings():
    """Return the Biolog mappings internally used by gempipe.
    
    Plate information is taken from DuctApe (https://doi.org/10.1016/j.ygeno.2013.11.005).
        
    Returns: 
        pandas.DataFrame: the Biolog mappings.
    """
    
    with resources.path('gempipe.assets', 'biolog_mappings.csv' ) as asset_path: 
        biolog_mappings = pnd.read_csv(asset_path, index_col=0)
    
    return biolog_mappings



def add_demand(model, mid):
    """Create a demand reaction, useful for debugging models.
    
    Args:
        model (cobra.Model): target model.
        mid (str): metabolite ID (compartment included) for which to create the demand.
        
    Returns:
        str: demand reaction ID.
    """
    
    rid = f"demand_{mid}"
    newr = cobra.Reaction(rid)
    model.add_reactions([newr])
    model.reactions.get_by_id(rid).reaction = f"{mid} -->"
    model.reactions.get_by_id(rid).bounds = (0, 1000)
    
    return rid



def can_synth(model, mid):
    """Check if the model can synthesize a given metabolite.
    
    Args:
        model (cobra.Model): target model.
        mid (str): metabolite ID (compartment included) for which to check the synthesis.
    
    Returns:
        (bool, float, str):
        
            `[0]` ``True`` if `mid` can be synthesized (``optimal`` status and positive flux).
        
            `[1]` maximal theoretical synthesis flux.
            
            `[2]` status of the optimizer.
    """
    
    # changes are temporary: demand is not added, objective is not changed.
    with model: 

        rid = add_demand(model, mid)

        # set the objective to this demand reaction:
        model.objective = rid

        # perform FBA: 
        res = model.optimize()
        value = res.objective_value
        status = res.status
        response = True if (value > 0 and status == 'optimal') else False
        
        return response, round(value, 2), status
    
    
    
def check_reactants(model, rid, verbose=True):
    """Check which reactant of a given reaction cannot be synthesized.
    
    Args:
        model (cobra.Model): target model.
        rid (str): reaction ID for which to check the synthesis of the reactants.
        verbose (bool): if False, just return the list of metabolite IDs without printing any further information.

    
    Returns:
        list: IDs of blocked reactants.
    """
    
    # changes are temporary
    with model: 
        counter = 0

        
        # get reactants and products
        reacs = [m for m in model.reactions.get_by_id(rid).reactants]
        prods = [m for m in model.reactions.get_by_id(rid).products]

        
        # iterate through the reactants: 
        mid_blocked = []
        for i, m in enumerate(reacs):
            
            # check if it can be synthesized:
            response, flux, status = can_synth(model, mid=m.id)
            
            if response==False: 
                counter += 1
                if verbose: print(f'{counter} : {flux} : {status} : {m.id} : {m.name}')
                mid_blocked.append(m.id)

                
        return mid_blocked
    
    
    
def sensitivity_analysis(model, scaled=False, top=3, mid=None):
    """Perform a sensitivity analysis (or reduced costs analysis) focused on the EX_change reaction.
    
    It is based on the current model's objective. The returned dictionary is sorted from most negative to most positive values.
    
    Args:
        model (cobra.Model): target model.
        scaled (bool): whether to scale to the current objective value.
        top (int): get just the first and last `top` EX_change reactions. If ``None``, all EX_change reactions will be returned.
        mid (str): instead of optimizing for the current objective reaction, do the analysis on the biosynthesis of a specific metabolite having ID `mid`. If `None` it will be ignored.
    
    Returns:
        dict: reduced costs keyd by EX_change reaction ID. 
    """
    
    
    # temporary chenges:
    with model:
        
      
        # focus on a specific metbaolite
        if mid != None: 
            model.objective = add_demand(model, mid)


        res = model.optimize()
        obj = res.objective_value
        flx = res.fluxes.to_dict()
        rcs = res.reduced_costs.to_dict()


        # manage 0 flux exception:
        if obj == 0 and scaled == True:
            raise Exception("Cannot scale reduced costs id the objective value is 0")


        # get the reduced costs of the EXR only:
        rcsex = {} 
        for key in rcs:
            if key.startswith("EX_"):
                if not scaled : rcsex[key] = rcs[key]
                else: rcsex[key] = rcs[key] * flx[key] / obj


        # get the most impactful (lowest and highest)
        rcsex = sorted(rcsex.items(), key=lambda x: x[1], reverse=True)
        rcsex = {i[0]: i[1] for i in rcsex}  # convert list to dictionary
    
        
        # get only the top N and bottom N exchanges
        if top != None: 
            rcsex_filt = {}
            for i in range(top):
                rcsex_filt[ list(rcsex.keys())[i] ] = list(rcsex.values())[i]
            for i in range(top):
                rcsex_filt[ list(rcsex.keys())[-top +i] ] = list(rcsex.values())[-top +i]
            rcsex = rcsex_filt
        
        return rcsex
    
    
    
def query_pam(name=[], ko=[], kr=[], km=[], kt=[], ec=[], des=[], pfam=[], annot=False, model=None, rid=None):
    """Show clusters in the context of a PAM.
    
    Clusters can be selected based on thier functional annotation.
    
    Args: 
        name (list): preferred names to search for, eg ['fabB', 'fabG'].
        ko (list): KEGG Orthologs (KOs) to search for, eg ['K00647', 'K00059'].
        kr (list): KEGG Reaction IDs to search for, eg ['R03460'].
        km (list): KEGG Module IDs to search for, eg ['M00846'].
        kt (list): KEGG Transport IDs to search for, eg ['3.A.3.2'].
        ec (list): EC codes to search for, eg ['2.5.1.19'].
        des (list): gene function descriptions to search for.
        pfam (list): PFAM domains to search for, eg ['Pyridox_ox_2'].
        annot (bool): if ``True``, return the functional annotation table instead of the PAM. 
        model (cobra.Model): if not ``None``, create a new column `modeled` as first, revealing the presence of each cluster in the model.
        rid (list): show clusters involved in the specified reactions (provided their IDs). Requires ``model``.
    
    Returns:
        pandas.DataFrame: filtered PAM or annotation table. 
    """
    
    
    # get the needed file from the cache:
    global __GAPSCACHE__
    if __GAPSCACHE__ == None:
        print("ERROR: you first need to execute gempipe.initialize().")
        return
    annotation = __GAPSCACHE__['annot']
    pam = __GAPSCACHE__['pam']
    report = __GAPSCACHE__['report']

    
    # create a copy to filter: 
    annotation_filter = annotation.copy()
    cluster_set = set(annotation_filter.index.to_list())
    
    
    # filter based on involved clusters: 
    if model != None: 
        if rid != None: 
            if type(rid)==str: rid = [rid]
            good_clusters = []
            for i in rid: 
                r = model.reactions.get_by_id(i)
                good_clusters = good_clusters + [g.id for g in r.genes if g.id in annotation.index]
            cluster_set = cluster_set.intersection(set(good_clusters))
                
    
    # filter for kegg orthologs
    if ko != []:
        if type(ko)==str: ko = [ko]
        good_clusters = []
        for i in ko:
            good_clusters = good_clusters + list(annotation[annotation['KEGG_ko'].str.contains(f'ko:{i}')].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
       
    # filter for kegg reactions
    if kr != []:
        if type(kr)==str: kr = [kr]
        good_clusters = []
        for i in kr:
            good_clusters = good_clusters + list(annotation[annotation['KEGG_Reaction'].str.contains(f'{i}')].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
        
    # filter for KEGG Modules
    if km != []:
        if type(km)==str: km = [km]
        good_clusters = []
        for i in km:
            good_clusters = good_clusters + list(annotation[annotation['KEGG_Module'].str.contains(f'{i}')].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
        
    # filter for KEGG Transport
    if kt != []:
        if type(kt)==str: kt = [kt]
        good_clusters = []
        for i in kt:
            good_clusters = good_clusters + list(annotation[annotation['KEGG_TC'].str.contains(f'{i}', case=False)].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
    
    
    # filter for gene symbols
    if name != []:
        if type(name)==str: name = [name]
        good_clusters = []
        for i in name:
            good_clusters = good_clusters + list(annotation[annotation['Preferred_name'].str.contains(f'{i.lower()}', case=False)].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
        
    # filter for EC:
    if ec != []:
        if type(ec)==str: ec = [ec]
        good_clusters = []
        for i in ec:
            good_clusters = good_clusters + list(annotation[annotation['EC'].str.contains(f'{i}', case=False)].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
        
    # filter for pfam domains
    if pfam != []:
        if type(pfam)==str: pfam = [pfam]
        good_clusters = []
        for i in pfam:
            good_clusters = good_clusters + list(annotation[annotation['PFAMs'].str.contains(f'{i}', case=False)].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
        
    # filter for function description
    if des != []:
        if type(des)==str: des = [des]
        good_clusters = []
        for i in des:
            good_clusters = good_clusters + list(annotation[annotation['Description'].str.contains(f'{i}', case=False)].index)
        cluster_set = cluster_set.intersection(set(good_clusters))
        
    
    # get tabular results (PAM or annotation)
    annotation_filter = annotation_filter.loc[list(cluster_set), ]
    if annot: results = annotation_filter
    else: results = pam.loc[[i for i in annotation_filter.index if i in pam.index], ]
    
    
    # rename columns using the species information:
    if not annot:
        old_to_new = {}
        for column in results.columns:
            match = report[report['accession']==column].iloc[0]
            old_to_new[column] = f"{match['species']} {match['accession']}"
        results = results.rename(columns=old_to_new)  # rename columns.
        results = results.reindex(sorted(results.columns), axis=1)  # alphabetical order.
    
    
    # mark clusters that are already modeled:
    if model != None: 
        results_columns = list(results.columns)
        results['modeled'] = 'False'
        for cluster in results.index:
            if cluster in [g.id for g in model.genes]:
                results.loc[cluster, 'modeled'] = ', '.join([r.id for r in model.genes.get_by_id(cluster).reactions])
        results = results[['modeled'] + results_columns]  # reorder columns
    
    
    return results



def import_from_universe(model, universe, rid, bounds=None, gpr=None):
    """Insert a new reaction taken from a universe model.

    Args:
        model (cobra.Model): target model to expand with new reactions.
        universe (cobra.Model): universe model, source of new reactions.
        rid (str): Id of the reaction to transfer. 
        bounds (tuple): bounds to apply to the inserted reaction, eg (0, 1000). If ``None``, bounds from universe will be retained.
        gpr (str): GPR to associate to the inserted reaction. If ``None``, no GPR will be associated.
    
    """
    
    r = universe.reactions.get_by_id(rid)
    model.add_reactions([r])
    
    if bounds != None:
        model.reactions.get_by_id(rid).bounds = bounds
        
    if gpr != None:
        model.reactions.get_by_id(rid).gene_reaction_rule = gpr
        model.reactions.get_by_id(rid).update_genes_from_gpr()
        
        if gpr == '':
            # be shure we are not importing panmodel genes:
            to_remove = []
            involved_gids = [g.id for g in model.genes]
            for g in model.genes: 
                if g.id in involved_gids and len(g.reactions)==0:
                    to_remove.append(g)
            cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)



def ss_preview(panmodel, accession):
    """Create a preview of a strain-specific model given its accession and the starting draft pan-GSMM.
    
    Args:
        panmodel (cobra.Model): draft pan-GSMM (ideally undergoing manual curation) from which to derive the strain-specific model.
        accession (str): accession of the strain for which to derive the strain-specific model.
        
    Returns:
        cobra.Model: strain-specific metabolic model.
        
    """
    
    # import here to avoid the "circular import" error:
    from gempipe.derive.strain import subtract_clusters
    
    
    # get the needed file from the cache:
    global __GAPSCACHE__
    if __GAPSCACHE__ == None:
        print("ERROR: you first need to execute gempipe.initialize().")
        return
    annotation = __GAPSCACHE__['annot']
    pam = __GAPSCACHE__['pam']
    report = __GAPSCACHE__['report']
    
    
    # create strain specific model
    ss_model = panmodel.copy()  
    response = subtract_clusters(ss_model, panmodel, pam, accession)
    
    
    return ss_model



def biolog_preview(model, starting_C=None, starting_N=None, starting_P=None, starting_S=None, seed=False):
    """Get a preview of of the Biolog simulations for the provided model.
    
    Args:
        model (cobra.Model): target model.
        starting_C (str): starting C source; if `None`, will be assigned to `EX_glc__D_e` or `EX_cpd00027_e0` depending on the selected database.
        starting_N (str): starting N source; if `None`, will be assigned to `EX_nh4_e` or `EX_cpd00013_e0` depending on the selected database.
        starting_P (str): starting P source; if `None`, will be assigned to `EX_pi_e` or `EX_cpd00009_e0` depending on the selected database.
        starting_S (str): starting S source; if `None`, will be assigned to `EX_so4_e` or `EX_cpd00048_e0` depending on the selected database.
        seed (bool): if `True`, use the SEED notation instead of the BiGG notation.
        
    Returns:
        pandas.DataFrame: Biolog simulations.
        
    """
    
    # import here to avoid the "circular import" error:
    from gempipe.derive.biolog import biolog_simulation
    
    
    if starting_C==None: 
        if seed==True: starting_C = 'EX_cpd00027_e0'
        else: starting_C = 'EX_glc__D_e'
    if starting_N==None: 
        if seed==True: starting_N = 'EX_cpd00013_e0'
        else: starting_N = 'EX_nh4_e'
    if starting_P==None: 
        if seed==True: starting_P = 'EX_cpd00009_e0'
        else: starting_P = 'EX_pi_e'
    if starting_S==None: 
        if seed==True: starting_S = 'EX_cpd00048_e0'
        else: starting_S = 'EX_so4_e'
    
    # do the simulations
    biolog_mappings = get_biolog_mappings()
    results_df = biolog_simulation(model, biolog_mappings, seed, starting_C, starting_N, starting_P, starting_S)
    
    
    return results_df



def add_reaction(model, rid, rstring, bounds=None, name='', gpr=None, ):
    """Add a new reaction to the model.
    
    Useful for example when the universe needs to be expanded.
    
    Args:
        model (cobra.Model): target model to expand with the new reaction.
        rid (str): ID for the new reaction.
        rstring (str): string describing the reaction. 
        bounds (set): lower and upper bound for the new reaction. 
        name (str): a name for the new reaction. 
        gpr (str): string describing the GPR.
        
    """
    
    r = cobra.Reaction(rid)
    r.name = name
    model.add_reactions([r])
    r = model.reactions.get_by_id(rid)
    r.build_reaction_from_string(rstring)
    if bounds != None: 
        r.bounds = bounds
    if gpr != None: 
        r.gene_reaction_rule = gpr
        r.update_genes_from_gpr()
        
        

def add_metabolite(model, mid, formula=None, charge=None, name=''):
    """Add a new metabolite to the model.
    
    Useful for example when the universe needs to be expanded.
    
    Args:
        model (cobra.Model): target model to expand with the new metabolite.
        mid (str): ID for the new metabolite.
        formula (str): string the chemical formula.
        charge (int): charge of the metabolite. 
        name (str): a name for the new metabolite. 
        
    """
    
    m = cobra.Metabolite(mid)
    m.name = name
    model.add_metabolites([m])
    m = model.metabolites.get_by_id(mid)
    m.formula = formula
    m.charge = charge
    m.compartment = mid.rsplit('_')[-1]
        


def search_similar(panmodel, rid, field='ko', unmod=False, species=None, showgpr=False, forceshow=False):
    """Search the PAM for gene clusters having functional annotation similar to that of gene clusters involved in the specified reaction of the pan-GSMM.
    
    First, the reaction `rid` is extracted from the `panmodel`. 
    Then, gene clusters are extracted from its GPR.
    The PAM is finally searched for all the gene clusters having the same functional annotation in one of the following fields: 'ko', 'pfam', 'kt', and 'ec'.
    
    Args:
        panmodel (cobra.Model): a pan-GSMM having the reaction `rid`.
        rid (str): ID of the reaction using the gene clusters on which to focus the search.
        field (str): functional annotation field to be used to extract similar gene clusters.
        unmod (bool): if ``True``, show only gene clusters that are not yet part of the `panmodel` (unmodeled).
        species (str): show only columns (genomes) having this substring in thier name (useful to filter for particular species or strains).
        showgpr (bool): if ``True``, show the pan-GSMM GPR of the selected reaction `rid`.
        forceshow (bool): if ``True``, display the results table with no limits on the number of displayable rows and columns. 
    
    Returns:
        pandas.DataFrame: PAM-derived results table, mixing presence/absence of gene clusters together with their functional annotation.

    """
    
    # check the selected feat: 
    available_fields = ['ko', 'pfam', 'kt', 'ec']
    if field not in available_fields: 
        print(f"field not recognized. Please use one of {available_fields}.")
        return
    results = None
    
    
    # show GPR in panmodel
    if showgpr: 
        print(f'GPR for {rid}:', panmodel.reactions.get_by_id(rid).gene_reaction_rule)
    involved_genes = [g.id for g in panmodel.reactions.get_by_id(rid).genes]
    
    
    if field == 'pfam': 
        terms = list(set(query_pam(annot=True).loc[involved_genes, ]['PFAMs'].to_list()))
        if terms == ['-']: return None
        presence_absence = query_pam(pfam=terms, model=panmodel)
        func_annot = query_pam(pfam=terms, annot=True)[['Description', 'KEGG_ko', 'PFAMs']]
        
        
    if field == 'ko': 
        terms = list(set(query_pam(annot=True).loc[involved_genes, ]['KEGG_ko'].to_list()))
        if terms == ['-']: return None
        terms = [i.replace('ko:', '') for i in terms]
        presence_absence = query_pam(ko=terms, model=panmodel)
        func_annot = query_pam(ko=terms, annot=True)[['Description', 'KEGG_ko', 'PFAMs']]
        
        
    if field == 'kt': 
        terms = list(set(query_pam(annot=True).loc[involved_genes, ]['KEGG_TC'].to_list()))
        if terms == ['-']: return None
        presence_absence = query_pam(kt=terms, model=panmodel)
        func_annot = query_pam(kt=terms, annot=True)[['Description', 'KEGG_TC', 'PFAMs']]
        
        
    if field == 'ec': 
        terms = list(set(query_pam(annot=True).loc[involved_genes, ]['EC'].to_list()))
        if terms == ['-']: return None
        presence_absence = query_pam(ec=terms, model=panmodel)
        func_annot = query_pam(ec=terms, annot=True)[['Description', 'EC', 'KEGG_ko', 'PFAMs']]
            
    
    # filter for desired columns: 
    if species != None: 
        presence_absence = presence_absence[['modeled']+[i for i in presence_absence.columns if species in i]]

    
    # merge presence/absence table and functional annootation table:
    results = pnd.concat([presence_absence, func_annot], axis=1)
    
    
    # exclude gene clusters already modeled: 
    if unmod: 
        results = results[results['modeled']=='False']
    
    
    # compute percentage of genomes with no representative (among this selection of gene clusters)
    empty_columns = results.columns[results.isnull().all()]
    structure_cols = ['modeled', 'Description', 'EC', 'KEGG_ko', 'PFAMs', 'KEGG_TC']  # cols not related to input genomes
    empty_columns = set(empty_columns) - set(structure_cols)
    all_columns = set(list(results.columns)) - set(structure_cols)
    print(f'Empty columns: {len(empty_columns)}/{len(all_columns)} ({round(len(empty_columns)/len(all_columns)*100,1)}%); {field} terms considered: {str(terms)}.')

    
    if forceshow:   # temporary allow pandas/jupyter to display with no row/col limits: 
        with pnd.option_context('display.max_rows', None, 'display.max_columns', None):
            display(results)
    
    return results


import os
import pickle
import multiprocessing
import itertools
import shutil
import subprocess
import warnings


import pandas as pnd
import cobra
from cobra.util.solver import linear_reaction_coefficients
from Bio import SeqIO, SeqRecord, Seq


from ..commons import get_blast_header
from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import get_retained_accessions
from ..commons import read_refmodel



def task_brh(proteome, args):
    
    
    # retrive the arguments:
    ref_proteome = args['ref_proteome']
    
    
    # get the basename without extension:
    basename = os.path.basename(proteome)
    accession, _ = os.path.splitext(basename)
    
    
    # create subdir without overwriting: 
    os.makedirs(f'working/brh/{accession}/', exist_ok=True)
    os.makedirs(f'working/brh/{accession}/dbs/reference/', exist_ok=True)
    os.makedirs(f'working/brh/{accession}/dbs/{accession}/', exist_ok=True)
            
    
    # create blast database for reference: 
    shutil.copyfile(ref_proteome, f'working/brh/{accession}/dbs/reference/ref_proteome.faa')  # just the content, not the permissions.  
    command = f"""makeblastdb -in working/brh/{accession}/dbs/reference/ref_proteome.faa -dbtype prot"""
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()

    
    # create blast database for current strain:
    shutil.copyfile(proteome, f'working/brh/{accession}/dbs/{accession}/{accession}.faa')  # just the content, not the permissions.    
    command = f"""makeblastdb -in working/brh/{accession}/dbs/{accession}/{accession}.faa -dbtype prot"""
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()
    

    # perform blastp for reference-on-accession: 
    command = f'''blastp \
        -query working/brh/{accession}/dbs/reference/ref_proteome.faa \
        -db working/brh/{accession}/dbs/{accession}/{accession}.faa \
        -out working/brh/{accession}/align_ref_vs_acc.tsv \
        -outfmt "6 {get_blast_header()}"
    '''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    # perform blastp for accession-on-reference: 
    command = f'''blastp \
        -query working/brh/{accession}/dbs/{accession}/{accession}.faa \
        -db working/brh/{accession}/dbs/reference/ref_proteome.faa \
        -out working/brh/{accession}/align_acc_vs_ref.tsv \
        -outfmt "6 {get_blast_header()}"
    '''
    process = subprocess.Popen(command, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    process.wait()


    # read both the alignments:
    header = f"{get_blast_header()}".split(' ')
    align_ref_vs_acc = pnd.read_csv(f'working/brh/{accession}/align_ref_vs_acc.tsv', names=header, sep='\t')
    align_ref_vs_acc['qcov'] = round((align_ref_vs_acc['qend'] -  align_ref_vs_acc['qstart'] +1)/ align_ref_vs_acc['qlen'] * 100, 1)
    align_acc_vs_ref = pnd.read_csv(f'working/brh/{accession}/align_acc_vs_ref.tsv', names=header, sep='\t')
    align_acc_vs_ref['qcov'] = round((align_acc_vs_ref['qend'] -  align_acc_vs_ref['qstart'] +1)/ align_acc_vs_ref['qlen'] * 100, 1)
    
    
    # save the alignments in a more readible formt:
    align_ref_vs_acc.to_csv(f'working/brh/{accession}_align_ref_vs_acc.csv')
    align_acc_vs_ref.to_csv(f'working/brh/{accession}_align_acc_vs_ref.csv')
    

    # parse the alignments
    results_df = [] 
    for cds in align_acc_vs_ref['qseqid'].unique():

        
        # acc_vs_ref
        curr_hsps = align_acc_vs_ref[align_acc_vs_ref['qseqid'] == cds]
        curr_hsps_filt = curr_hsps[(curr_hsps['qcov'] >= 70) & (curr_hsps['evalue'] <= 1e-5)]
        curr_hsps_filt_sort = curr_hsps_filt.sort_values(by='evalue', ascending=True)
        curr_hsps_filt_sort.reset_index(inplace=True, drop=True)
        try: best_hit = curr_hsps_filt_sort.loc[0, 'sseqid']
        except: 
            results_df.append({'cds': cds, 'ref': None, 'reciprocal': 'NA'})
            continue


        # ref_vs_acc
        curr_hsps2 = align_ref_vs_acc[align_ref_vs_acc['qseqid'] == best_hit]
        curr_hsps_filt2 = curr_hsps2[(curr_hsps2['qcov'] >= 70) & (curr_hsps2['evalue'] <= 1e-5)]
        curr_hsps_filt_sort2 = curr_hsps_filt2.sort_values(by='evalue', ascending=True)
        curr_hsps_filt_sort2.reset_index(inplace=True, drop=True)
        try: best_hit2 = curr_hsps_filt_sort2.loc[0, 'sseqid']
        except: 
            results_df.append({'cds': cds, 'ref': best_hit, 'reciprocal': 'NA'})
            continue

        
        # annotate if bi-directional or mono-directional
        if cds == best_hit2: results_df.append({'cds': cds, 'ref': best_hit, 'reciprocal': '<=>'})
        else: results_df.append({'cds': cds, 'ref': best_hit, 'reciprocal': '=>'})


    # save results to disk
    ref_proteome_basename = os.path.basename(ref_proteome)
    results_df = pnd.DataFrame.from_records(results_df)
    results_df.to_csv(f'working/brh/{accession}_brh_{ref_proteome_basename}.csv')
    
    
    # save disk space removeing databases: 
    shutil.rmtree(f'working/brh/{accession}/') 
    
    
    # return a row for the dataframe
    return [{'accession': accession, 'completed': True}]
    


def create_refgid_to_clusters(logger, refmodel_basename, ref_proteome_basename, edits_dict): 
    
    
    # load the previously created doctionaries: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
    with open('working/clustering/seq_to_cluster.pickle', 'rb') as handler:
        seq_to_cluster = pickle.load(handler)
        
        
    # create a 'cds_to_newcluster' dict, parsing the edits_dict.
    # This is because one of the ss proteins in the brh results coul be 'broken'.
    # Being broken, its cluster would be changed.
    cds_to_newcluster = {}
    if edits_dict != None:
        # 'edits_dict' syntax is "x:cluster --> frag_x_y:newcluster"
        for key, value in edits_dict.items():
            cds = key.split(':', 1)[0]
            new_cluster = value.split(':', 1)[1]
            if cds in cds_to_newcluster.keys():
                logger.error("Each 'cds' should appear just one time in 'create_refgid_to_clusters()'.")
            cds_to_newcluster[cds] = new_cluster    
    # save the dictionary: 
    with open(f'working/brh/cds_to_newcluster.pickle', 'wb') as handler:
        pickle.dump(cds_to_newcluster, handler)
    
    
    # parse each filtered accession: 
    refgid_to_clusters = {}
    for species in species_to_proteome.keys():
        for proteome in species_to_proteome[species]: 
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
    
            
            # read the brh results for this accession: 
            df_result = pnd.read_csv(f'working/brh/{accession}_brh_{ref_proteome_basename}.csv', index_col=0)
            df_result = df_result.set_index('cds', drop=True, verify_integrity=True)
            
            
            # get only the reciprocal hits: 
            df_brh = df_result[df_result['reciprocal'] == '<=>']
            
            
            # populate the dictionary: 
            for cds, row in df_brh.iterrows(): 
                if row['ref'] not in refgid_to_clusters.keys(): 
                    refgid_to_clusters[row['ref']] = set()
                    
                    
                # take into account eventual 'broken' proteins
                cluster = seq_to_cluster[cds]
                if edits_dict != None:
                    if cds in cds_to_newcluster.keys():
                        cluster = cds_to_newcluster[cds]
                    
                # finally add the cluster
                refgid_to_clusters[row['ref']].add(cluster)


    # save the dictionary: 
    with open(f'working/brh/{refmodel_basename}.refgid_to_clusters.pickle', 'wb') as handler:
        pickle.dump(refgid_to_clusters, handler)
        
        
        
def translate_refmodel(logger, refmodel, ref_proteome, refspont): 
    
    
    # load the model according to the file type
    logger.info("Loading the provided reference model...")
    refmodel_basename = os.path.basename(refmodel)
    refmodel = read_refmodel(refmodel)   # handle variuos formats
    if type(refmodel)==int: return 1  # an error was raised
    logger.info(f"Done, {' '.join(['G:', str(len(refmodel.genes)), '|', 'R:', str(len(refmodel.reactions)), '|', 'M:', str(len(refmodel.metabolites))])}.")
    
    
    # print the preloaded objective reaction
    objs = list(linear_reaction_coefficients(refmodel).keys())
    if len(objs) > 1: logger.warning("More than 1 objective reactions were set up. Showing the first.")
    logger.info(f"The following objective was set up: {objs[0].id}.")
    
    
    # save the reference model in a standard format
    logger.debug("Saving a copy of the reference model in JSON format...")
    cobra.io.save_json_model(refmodel, f'working/brh/{refmodel_basename}.refmodel_original.json') # ext can be repeated.
    
    
    # create a copy, later translated
    logger.info("Converting reference model's gene notation to clusters...")
    refmodel_t = refmodel.copy()

    
    # get the modeled genes: 
    modeled_gids = set([g.id for g in refmodel.genes])


    # sometimes the reference proteome contains less genes respect to those modeled.
    # we remove from the reference model the genes missing from the proteome: 
    gids_in_proteome = set()
    with open(ref_proteome, 'r') as r_handler:                  
        for seqrecord in SeqIO.parse(r_handler, "fasta"):
            gid = seqrecord.id
            gids_in_proteome.add(gid)
    # remove non-proteome genes (exclude the "spontaneous" gene): 
    if len(modeled_gids - gids_in_proteome) > 0:
        to_remove = []
        for i in list(modeled_gids - gids_in_proteome):
            if i != refspont:  # exclude the "spontaneous" gene
                to_remove.append(i)
        logger.info(f"The following genes will be removed from the reference model, as they do not appear in the reference proteome: {to_remove}.") 
        to_remove = [refmodel_t.genes.get_by_id(gid) for gid in to_remove]
        cobra.manipulation.remove_genes(refmodel_t, to_remove, remove_reactions=False)
        # update 'modeled_gids' set:
        modeled_gids = modeled_gids - set([g.id for g in to_remove])


    # load the refgid_to_clusters dictionary (1-to-many)
    with open(f'working/brh/{refmodel_basename}.refgid_to_clusters.pickle', 'rb') as handler:
        refgid_to_clusters = pickle.load(handler)
    
        
    # if gempipe is run in 'reference' mode, on a gempipe-generated reference pan-model,
    # the the old gene IDs and current clusteres will be conflicting. In these cases,
    # we add a prefix to the old gene ID, and correct 'modeled_gids' and 'refgid_to_clusters' accordingly.
    gempipe_on_gempipe = False
    if any(['Cluster_' in g.id for g in refmodel_t.genes]):
        gempipe_on_gempipe = True  # flag
        renaming_dict = {}
        for g in refmodel_t.genes:
            if g.id == refspont:
                continue
                
            # edit the 'modeled_gids' set:
            modeled_gids = modeled_gids - set([g.id])
            modeled_gids.add(f'old_{g.id}')
            
            # edit the 'gids_in_proteome' set:
            gids_in_proteome = gids_in_proteome - set([g.id])
            gids_in_proteome.add(f'old_{g.id}')
            
            # edit the 'refgid_to_clusters' dict:
            if g.id in refgid_to_clusters.keys():
                pointed = refgid_to_clusters[g.id]
                del refgid_to_clusters[g.id]
                refgid_to_clusters[f'old_{g.id}'] = pointed
                
            # populate the 'renaming_dict' (old: new)
            renaming_dict[g.id] = f'old_{g.id}'
            
        # finally rename genes with their prefixed version:
        cobra.manipulation.rename_genes(refmodel_t, renaming_dict)
            
        
    # at this point, the 'refgid_to_clusters' dict may not contain some reference genes.
    # This is because of the threshold applied earlier, on the 'qcov' and 'evalue', for the determination of the brh.
    # Here we try to recovere these genes: 
    translable_gids = set(list(refgid_to_clusters.keys()))
    no_brh_gids = (modeled_gids - translable_gids) - set([refspont])
    if len(no_brh_gids) > 0:
        logger.info(f"The genes {no_brh_gids} remained without a BRH. Now starting their recovery.")
        recovered_translations = {}
        
        # load the translation dict (also take into account the 'broken' proteins)
        with open('working/clustering/seq_to_cluster.pickle', 'rb') as handler:
            seq_to_cluster = pickle.load(handler)
        with open(f'working/brh/cds_to_newcluster.pickle', 'rb') as handler:
            cds_to_newcluster = pickle.load(handler)
        
        # iterate the 'good' accessions
        with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
            species_to_proteome = pickle.load(handler)
        for species in species_to_proteome.keys():
            for proteome in species_to_proteome[species]: 
                basename = os.path.basename(proteome)
                accession, _ = os.path.splitext(basename)

                # read the brh results for this accession: 
                ref_vs_acc = pnd.read_csv(f'working/brh/{accession}_align_ref_vs_acc.csv', index_col=0)
                
                # get the best match for each 'gid':
                for gid in no_brh_gids:
                    if gempipe_on_gempipe: gid = gid[len('old_'):]
                    matches = ref_vs_acc[ref_vs_acc['qseqid']==gid]
                    matches = matches.sort_values(by='evalue', ascending=True)
                    matches = matches.reset_index(drop=True)
                    
                    try: match = matches.iloc[0]
                    except: continue  # no match
                    
                    # populate the dictionary:
                    if gempipe_on_gempipe: gid = f'old_{gid}'
                    if gid not in recovered_translations.keys():
                        recovered_translations[gid] = set()
                    cds = match['sseqid']
                    cluster = seq_to_cluster[cds]
                    if cds in cds_to_newcluster.keys():  # manage also the case of 'broken' proteins
                        cluster = cds_to_newcluster[cds]
                    recovered_translations[gid].add(cluster) 
                    
        logger.debug(f'Recovered translations: {recovered_translations}')
        # supplement the 'refgid_to_clusters' dict:
        for key, value in recovered_translations.items():
            refgid_to_clusters[key] = value
                    
                
    # finally rename the genes: 
    for r in refmodel_t.reactions:
        gpr = r.gene_reaction_rule
        
        
        # force each gid to be surrounded by spaces: 
        gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '

        
        # translate the spontaneus gene
        if f' {refspont} ' in gpr:
            gpr = gpr.replace(f' {refspont} ', f' spontaneous ')

        
        # translate this GPR
        for gid in refgid_to_clusters.keys():
            if f' {gid} ' in gpr:  
                gpr = gpr.replace(f' {gid} ', f' ({" or ".join(refgid_to_clusters[gid])}) ')


        # remove spaces between parenthesis
        gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
        # remove spaces at the extremes: 
        gpr = gpr[1: -1]


        # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
        r.gene_reaction_rule = gpr
        r.update_genes_from_gpr()


    # now remove the reference genes:
    to_remove = [g for g in refmodel_t.genes if g.id in gids_in_proteome]
    if refspont != 'spontaneous': to_remove = to_remove + [refmodel_t.genes.get_by_id(refspont)]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="need to pass in a list")
        cobra.manipulation.remove_genes(refmodel_t, to_remove, remove_reactions=True)
    logger.info(f"Done, {' '.join(['G:', str(len(refmodel_t.genes)), '|', 'R:', str(len(refmodel_t.reactions)), '|', 'M:', str(len(refmodel_t.metabolites))])}.")
    
    
    # save the reference model in a standard format
    logger.debug("Saving a copy of the converted reference model in JSON format...")
    cobra.io.save_json_model(refmodel_t, f'working/brh/{refmodel_basename}.refmodel_translated.json')  # ext can be repeated.
    
    
    return 0
    
    

def perform_brh(logger, cores, ref_proteome): 
    
    
    # some log messages:
    logger.info("Performing the best reciprocal hits (BRH) alignment against the reference proteome...")
    if not os.path.exists(ref_proteome): # check the input:
        logger.error(f"Provided path to the reference proteome (-rp/--ref_proteome) does not exist: {ref_proteome}.")
        return 1
    
    
    # create sub-directories without overwriting:
    os.makedirs('working/brh/', exist_ok=True)

    
    # load the previously created species_to_genome: 
    with open('working/proteomes/species_to_proteome.pickle', 'rb') as handler:
        species_to_proteome = pickle.load(handler)
    

    # check if it's everything pre-computed
    ref_proteome_basename = os.path.basename(ref_proteome)
    results_presence = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]:
            basename = os.path.basename(proteome)
            accession, _ = os.path.splitext(basename)
            results_presence.append(os.path.exists(f'working/brh/{accession}_brh_{ref_proteome_basename}.csv'))
            results_presence.append(os.path.exists(f'working/brh/{accession}_align_ref_vs_acc.csv'))
            results_presence.append(os.path.exists(f'working/brh/{accession}_align_acc_vs_ref.csv'))
    if all(results_presence): 
        # log some message: 
        logger.info('Found all the needed files already computed. Skipping this step.')
        # signal to skip this module:
        return 0
        

    # create items for parallelization: 
    items = []
    for species in species_to_proteome.keys(): 
        for proteome in species_to_proteome[species]: 
            items.append(proteome)
    

    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
    
    
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'completed']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_brh), 
            itertools.repeat({'ref_proteome': ref_proteome}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    
    
    return 0



def convert_reference(logger, refmodel, ref_proteome, gene_recovery, refspont):
    
    
    # some log messages:
    logger.info("Translating the reference model's genes to clusters...")
    if not os.path.exists(refmodel): # check the input:
        logger.error(f"Provided path to the reference model (-rm/--ref_model) does not exist: {refmodel}.")
        return 1
    
    
    # get basename for reference model:
    refmodel_basename = os.path.basename(refmodel)
    
    
    # check if it's everything pre-computed
    if os.path.exists('working/brh/proc_acc.pickle'):
        with open('working/brh/proc_acc.pickle', 'rb') as handler:
            proc_acc = pickle.load(handler) 
        if get_retained_accessions() == proc_acc:
            if os.path.exists(f'working/brh/{refmodel_basename}.refmodel_original.json'):
                if os.path.exists(f'working/brh/{refmodel_basename}.refmodel_translated.json'):
                    if os.path.exists(f'working/brh/{refmodel_basename}.refgid_to_clusters.pickle'):
                        # log some message: 
                        logger.info('Found all the needed files already computed. Skipping this step.')
                        # signal to skip this module:
                        return 0

    
    # create a dictionary ref_seq-to-clusters, parsing the BRHs. 
    ref_proteome_basename = os.path.basename(ref_proteome)
    if not gene_recovery: edits_dict = None  # load the 'edits_dict'
    else: edits_dict = pickle.load(open('working/rec_broken/edits_dict.pickle', 'rb'))
    create_refgid_to_clusters(logger, refmodel_basename, ref_proteome_basename, edits_dict)
    
    
    # get a copy of the refmodel, and translate its genes to clusters notation. 
    response = translate_refmodel(logger, refmodel, ref_proteome, refspont)
    if response == 1: return 1


    # make traces to keep track of the accessions processed:
    shutil.copyfile('working/annotation/proc_acc.pickle', 'working/brh/proc_acc.pickle')
    
    
    return 0
    


import os


import pandas as pnd
import cobra 


from .strain import derive_strain_specific
from .filler import strain_filler
from .biolog import strain_biolog_tests
from .species import derive_rpam
from .species import derive_species_specific



def derive_all(logger, outdir, cores, panmodel, pam, report, gannots, media_filepath, minflux, biolog, sbml, skipgf):
    
    
    ### PART 1: derive strain-specific models
    
    response = derive_strain_specific(logger, outdir, cores, panmodel, pam, report, gannots, sbml)
    if response == 1: return 1


    ### PART 2: gap-fill strain-specific models
    if skipgf: logger.info("Skipping the strain-specific gap-filling step (--skipgf)...")
    if not skipgf:
        response = strain_filler(logger, outdir, cores, panmodel, media_filepath, minflux, sbml)
        if response == 1: return 1
    

    ### PART 2bis: simulations
    if biolog:
        response = strain_biolog_tests(logger, outdir, cores, pam, panmodel, skipgf)
        if response == 1: return 1
    
    
    ### PART 3: derive species-specific models
    
    response = derive_rpam(logger, outdir, cores, panmodel, skipgf)
    if response == 1: return 1
    
    response = derive_species_specific(logger, outdir, cores, panmodel, sbml)
    if response == 1: return 1


    return 0



def derive_command(args, logger):
    
    
    # check the existence of the input files:
    if args.inpanmodel == '-' or args.inpam == '-' or args.inreport == '-':
        logger.error("Please specify the input pan-model (-im/--inpanmodel), PAM (-ip/--inpam) and report (-ir/--inreport).")
        return 1
    else:  # all 3 parameters were set
        if not os.path.exists(args.inpanmodel):
            logger.error(f"The specified path for input pan-model (-im/--inpanmodel) does not exist: {args.inpanmodel}.")
            return 1
        if not os.path.exists(args.inpam):
            logger.error(f"The specified path for input PAM (-ip/--inpam) does not exist: {args.inpam}.")
            return 1
        if not os.path.exists(args.inreport):
            logger.error(f"The specified path for input report (-ir/--inreport) does not exist: {args.inreport}.")
            return 1
        
        
    if args.ingannots != '-':
        if not os.path.exists(args.ingannots):
            logger.error(f"The specified path for input genes annotation table (-ig/--ingannots) does not exist: {args.ingannots}.")
            return 1
    
    
    # load input files
    logger.info("Loading input files...")
    panmodel = cobra.io.load_json_model(args.inpanmodel)
    pam = pnd.read_csv(args.inpam, index_col=0)
    report = pnd.read_csv(args.inreport, index_col=0)
    
    if args.ingannots != '-':
        # "low_memory=False" to prevent "DtypeWarning: Columns ... have mixed types."
        gannots = pnd.read_csv(args.ingannots, index_col=0, low_memory=False)
        gannots = gannots.astype(str)  # 'ncbigene' is still formatted as number
        gannots = gannots.replace('nan', None)
        #gannots['refseq'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in gannots['refseq'] ]
        gannots['ncbigene'] = [i.rsplit('.', 1)[0] if type(i)==str else None for i in gannots['ncbigene'] ]
    else: gannots = None
    
    
    # create the main output directory: 
    outdir = args.outdir
    if outdir.endswith('/') == False: outdir = outdir + '/'
    os.makedirs(outdir, exist_ok=True)
    
    
    logger.info("Deriving strain- and species-specific metabolic models...")
    response = derive_all(logger, outdir, args.cores, panmodel, pam, report, gannots, args.media, args.minflux, args.biolog, args.sbml, args.skipgf)
    if response == 1: return 1
    
    
    return 0



import multiprocessing
import itertools
from importlib import resources
import os
import warnings


import cobra
import pandas as pnd


from ..commons import chunkize_items
from ..commons import load_the_worker
from ..commons import gather_results
from ..commons import read_refmodel




def biolog_simulation(ss_model, biolog_mappings, seed=False, starting_C='EX_glc__D_e', starting_N='EX_nh4_e', starting_P='EX_pi_e', starting_S='EX_so4_e' ):
    
    
    # get the modeled rids: 
    modeled_rids = set([r.id for r in ss_model.reactions])
    
    
    # iterate each Biolog substrate
    df_results = []  # list of dicts
    for substrate, row in biolog_mappings.iterrows():
        # skip the substrate if 'source'==set():
        for source in eval(row['source']): 
            
            
            # get the 'substituting' and 'substitute' rids:
            exr_after = row['SEED_exchange'] if seed else row['BiGG_exchange']
            if source == 'C': exr_before = starting_C  # 'EX_glc__D_e'
            elif source == 'N': exr_before = starting_N  # 'EX_nh4_e'
            elif source == 'P': exr_before = starting_P  # 'EX_pi_e'
            elif source == 'S': exr_before = starting_S  # 'EX_so4_e'
            else: exr_before = '?'

            # flags for presence of reaction IDs:
            exr_after_present = exr_after in modeled_rids
            exr_before_present = exr_before in modeled_rids
            
            # other columns to fill
            formula = None
            growth = None
            value = None
            status = None
            

            # changes are temporary: 
            with ss_model:
                
                # perform the simulation:
                if exr_before_present:  # the exchange reaction could not be present
                    ss_model.reactions.get_by_id(exr_before).bounds = (0, 1000)
                    
                    # particularly important in case of rich media
                    res_before = ss_model.optimize()
                    objv_before = res_before.objective_value
                    if res_before.status == 'infeasible': objv_before = 0
                    
                    if exr_after_present: # the exchange reaction could not be present
                        r = ss_model.reactions.get_by_id(exr_after)
                        r.bounds = (-1000, 1000)
                        formula = list(r.metabolites)[0].formula
                        
                        # optimize catching the warnings
                        with warnings.catch_warnings(record=True) as w:
                            res = ss_model.optimize()
                        for warning in w:
                            pass
                        
                        value = res.objective_value
                        status = res.status
                        
                        growth = value >= (objv_before + 0.001) and status == 'optimal'
            
            
            # even if they could be simulated, substrate out of the Biolog system
            # are not taken into acocunt in this analysis:
            if eval(row['PM'])==set(): 
                growth = value = status = None
            

            # populate results dictionary:
            df_results.append({
                'substrate': substrate, 'source': source, 'formula': formula,  
                'exr_before': exr_before, 'exr_before_present': exr_before_present,
                'exr_after': exr_after, 'exr_after_present': exr_after_present,
                'growth': growth, 'value': value, 'status': status})
    df_results  = pnd.DataFrame.from_records(df_results)
    
    
    return df_results

    

def task_biolog(accession, args):
    
    
    # retrive the arguments:
    biolog_mappings = args['biolog_mappings']
    outdir = args['outdir']
    skipgf = args['skipgf']
    
    # read json/sbml file:
    if not skipgf:
        ss_model = cobra.io.load_json_model(outdir + f'strain_models_gf/{accession}.json')
    else:  # user asked to skip the strain-specific gapfilling step
        ss_model = cobra.io.load_json_model(outdir + f'strain_models/{accession}.json')
    
    # perform the simulations: 
    outfile = outdir + 'biolog/' + accession + '.csv'
    df_results = biolog_simulation(ss_model, biolog_mappings)
    df_results.to_csv(outfile)
    
 
    return [{'accession': accession, 'completed': True}]



def strain_biolog_tests(logger, outdir, cores, pam, panmodel, skipgf):
    
    
    # log some messages
    logger.info("Simulating Biolog's substrates utilization...")
    
    
    # create needed subdirectories:
    os.makedirs(outdir + 'biolog/', exist_ok=True)
    
    
    # check if it's everything pre-computed:
    # not needed!
    
    
    # load the assets to form the args dictionary:
    biolog_mappings = None
    with resources.path('gempipe.assets', 'biolog_mappings.csv' ) as asset_path: 
        biolog_mappings = pnd.read_csv(asset_path, index_col=0)
        
        
    # first of all, test the capabilities of the panmodel: 
    outfile = outdir + 'biolog_panmodel.csv'
    df_results = biolog_simulation(panmodel, biolog_mappings)
    df_results.to_csv(outfile)

    
    # create items for parallelization: 
    items = []
    for accession in pam.columns:
        items.append(accession)
       
        
    # randomize and divide in chunks: 
    chunks = chunkize_items(items, cores)
                          
                          
    # initialize the globalpool:
    globalpool = multiprocessing.Pool(processes=cores, maxtasksperchild=1)
    
    
    # start the multiprocessing: 
    results = globalpool.imap(
        load_the_worker, 
        zip(chunks, 
            range(cores), 
            itertools.repeat(['accession', 'completed']), 
            itertools.repeat('accession'), 
            itertools.repeat(logger), 
            itertools.repeat(task_biolog),  # will return a new sequences dataframe (to be concat).
            itertools.repeat({'biolog_mappings': biolog_mappings, 'outdir': outdir, 'skipgf': skipgf}),
        ), chunksize = 1)
    all_df_combined = gather_results(results)
    
    
    # empty the globalpool
    globalpool.close() # prevent the addition of new tasks.
    globalpool.join() 
    


import os
import shutil
import subprocess
import pickle


import pandas as pnd


import cobra


from .getgenomes import download_genomes
from .getgenomes import handle_manual_genomes
from .extractcds import extract_cds
from .extractcds import handle_manual_proteomes
from .handlegbff import handle_manual_genbanks
from .filtergenomes import filter_genomes
from .clustercds import compute_clusters
from .rec_masking import recovery_masking
from .rec_broken import recovery_broken
from .rec_overlap import recovery_overlap
from .funcannot import func_annot
from .networkrec import network_rec
from .tcdbing import tcdbing_main
from .reciprocalhits import perform_brh
from .reciprocalhits import convert_reference
from .refexpansion import ref_expansion
from .pimp import denovo_annotation
from .duplicates import solve_duplicates
from .reporting import create_panmodel_proteome
from .reporting import create_report
from .reporting import create_recon_plots


from ..commons import get_md5_string
from ..commons import get_outdir



def preliminary_checks(args, logger):
    
    
    # overwrite if requested:
    if os.path.exists('working/'):
        logger.info("Found a previously created ./working/ directory.")
        if args.overwrite:
            logger.info("Ereasing the ./working/ directory as requested (--overwrite).")
            shutil.rmtree('working/')
    os.makedirs('working/', exist_ok=True)
    os.makedirs('working/logs/', exist_ok=True)
    
    
    # check if the user required the list of databases: 
    if args.buscodb == 'show': 
        logger.info("Creating the temporary ./busco_downloads/ directory...")
        command = f"""busco --list-datasets"""
        process = subprocess.Popen(command, shell=True)
        process.wait()
        shutil.rmtree('busco_downloads/') 
        logger.info("Deleted the temporary ./busco_downloads/ directory.")
        return 0
        
    
    # check inputted gram staining 
    if args.staining != 'pos' and args.staining != 'neg': 
        logger.error("Gram staining (-s/--staining) must be either 'pos' or 'neg'.")
        return 1



def draft_reconstruction(args, logger):
    
    
    # assure the output directory
    outdir = get_outdir(args.outdir)
    
    
    ### PART 1. Obtain the preoteomes. 
    
    if args.genbanks != '-':
        # handle the manually defined genbanks: 
        response = handle_manual_genbanks(logger, args.genbanks, outdir)
        if response == 1: return 1
    
    elif args.proteomes != '-':
        # handle the manually defined proteomes: 
        response = handle_manual_proteomes(logger, args.proteomes)
        if response == 1: return 1
    
    elif args.genomes != '-':
        # handle the manually defined genomes: 
        response = handle_manual_genomes(logger, args.genomes)
        if response == 1: return 1
    
        # extract the CDSs from the genomes:
        response = extract_cds(logger, args.cores, outdir, args.nofig)
        if response == 1: return 1    
    
        # filter the genomes based on technical/biological metrics:
        response = filter_genomes(logger, args.cores, args.buscodb, args.buscoM, args.buscoF, args.ncontigs, args.N50, outdir, args.nofig)
        if response == 1: return 1  
    
    elif args.taxids != '-':
        # download the genomes according to the specified taxids: 
        response = download_genomes(logger, args.taxids, args.cores)
        if response == 1: return 1 
    
        # extract the CDSs from the genomes:
        response = extract_cds(logger, args.cores, outdir, args.nofig)
        if response == 1: return 1 
    
        # filter the genomes based on technical/biological metrics:
        response = filter_genomes(logger, args.cores, args.buscodb, args.buscoM, args.buscoF, args.ncontigs, args.N50, outdir, args.nofig)
        if response == 1: return 1  
    
    else:
        logger.error("Please specify the input species taxids (-t/--taxids) or the input genomes (-g/--genomes) or the input proteomes (-p/--proteomes) or the input genbanks (-gb/--genbanks).")
        return 1
    
    
    ### PART 2. Clustering. 
    
    # cluster the aminoacid sequences according to sequence similarity. 
    response = compute_clusters(logger, args.cores)
    if response == 1: return 1 


    ### PART 3. Gene recovery.
    gene_recovery = True
    if (args.proteomes != '-' or args.genbanks != '-') and args.norec == False:  # warning if starting from proteomes
        logger.warning("gempipe gives its best when starting from genomes. Starting from proteomes/genbanks will skip the gene recovery modules.")
        gene_recovery = False
        
    if args.norec:
        gene_recovery = False
    
    if gene_recovery: 
        # Recovery 1: search for proteins broken in two
        response = recovery_broken(logger, args.cores)
        if response == 1: return 1
        
        # Recovery 2: search missing genes after masking the genome 
        response = recovery_masking(logger, args.cores)
        if response == 1: return 1 
        
        # Recovery 3: search for overlapping genes
        response = recovery_overlap(logger, args.cores)
        if response == 1: return 1
    
    # define the final PAM in the current directory: 
    if gene_recovery:
        shutil.copyfile('working/rec_overlap/pam.csv', outdir + 'pam.csv')
    else: shutil.copyfile('working/clustering/pam.csv', outdir + 'pam.csv')
    
    
    ### PART 4. Reconstruction of the reference-free reaction network.
    
    # perform functional annotation
    response = func_annot(logger, args.cores, outdir, args.dbs, args.dbmem)
    if response == 1: return 1

    # define the final annotation in the current directory: 
    annotation = pnd.read_csv('working/annotation/pan.emapper.annotations', sep='\t', comment='#', header=None)
    annotation.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    annotation = annotation.set_index('query', drop=True, verify_integrity=True)
    annotation.to_csv(outdir + 'annotation.csv')

    # perform the reaction network reconstruction
    response = network_rec(logger, args.cores, args.staining, args.identity, args.coverage, args.refmodel, args.refproteome)
    if response == 1: return 1
    
    
    ### PART 5. Eventual reference-based reconstruction.
    
    if args.refmodel != '-' and args.refproteome != '-':
        
        # compute the best reciprocal hits for all the strains:
        response = perform_brh(logger, args.cores, args.refproteome)
        if response == 1: return 1
        
        # convert the reference model's genes to clusters: 
        response = convert_reference(logger, args.refmodel, args.refproteome, gene_recovery, args.refspont)
        if response == 1: return 1
    
        # expand the reference model with new reactions coming from the reference-free recon.
        response = ref_expansion(logger, args.refmodel, args.mancor, args.identity, args.coverage)
        if response == 1: return 1
    
    
    # define the final draft pan-model in the duplicates/ directory: 
    os.makedirs('working/duplicates/', exist_ok=True)  # without overwriting
    if args.refmodel != '-' and args.refproteome != '-':
        shutil.copyfile('working/expansion/draft_panmodel.json', 'working/duplicates/draft_panmodel.json')
    else: shutil.copyfile(f'working/free/draft_panmodel_{args.identity}_{args.coverage}.json', 'working/duplicates/draft_panmodel.json')
    
    
    ### PART X: Eventual TCDB-based transporters recon (experimental)
    
    if args.tcdb:  # experimental feature
        # try to reconstruct transport reactions using the tcdb. 
        response = tcdbing_main(logger, args.cores, args.staining)
        if response == 1: return 1
        
        
    return 0 



def automated_curation(args, logger):
    
    
    # assure the output directory: 
    outdir = get_outdir(args.outdir)
    
    
    # PART 6. Automated curation
    
    # perform de-novo annotation with metanetx 4.4 (aka pimp_my_model)
    response = denovo_annotation(logger)
    if response == 1: return 1


    #if args.refmodel != '-': 
    if args.dedup :
        # solve duplicate metabolites and reactions using mnx annotation
        response = solve_duplicates(logger, args.identity, args.coverage, args.refmodel, args.mancor)
        if response == 1: return 1

        # save the final panmodel
        shutil.copyfile(f'working/duplicates/draft_panmodel_da_dd.json', outdir + 'draft_panmodel.json')

    else:
        shutil.copyfile(f'working/duplicates/draft_panmodel_da.json', outdir + 'draft_panmodel.json')
    
        
    # PART 8. Create the reference panmodel proteome
    
    response = create_panmodel_proteome(logger, outdir)
    if response == 1: return 1
        
    
    # PART 7. Report
    
    # create a report table
    response = create_report(logger, outdir)
    if response == 1: return 1

    # create plots showing >preliminar< reconstruction metrics
    create_recon_plots(logger, outdir, args.cores, args.nofig)
    if response == 1: return 1

    
    return 0



def recon_command(args, logger):

    
    # overwrite, list databases, check staining ...
    response = preliminary_checks(args, logger)
    if response == 0: return 0
    if response == 1: return 1
    
    
    # produce a reference-free / reference-based reconstruction
    response = draft_reconstruction(args, logger)
    if response == 1: return 1


    # save the panmodel md5:
    panmodel_md5 = get_md5_string('working/duplicates/draft_panmodel.json')
    with open('working/duplicates/md5.pickle', 'wb') as handle: 
        pickle.dump(panmodel_md5, handle)
    
    
    # apply some automated curation (duplicates removal, ...)
    response = automated_curation(args, logger)
    if response == 1: return 1


    # make a SBML copy of the final draft panmodel if requested
    if args.sbml: 
        outdir = get_outdir(args.outdir)
        # load the final draft panmdoel
        draft_panmodel = cobra.io.load_json_model(outdir + 'draft_panmodel.json')
        # create a SBML copy
        cobra.io.write_sbml_model(draft_panmodel, outdir + 'draft_panmodel.xml')
    
    
    return 0
    
    
        
        
        
    
    


import os
from importlib import resources
import subprocess
import shutil
import pickle
import itertools


import pandas as pnd
from Bio import SeqIO
import cobra



from ..commons import get_blast_header
from ..commons import get_retained_accessions
from ..commons import read_refmodel


from ..curate.gaps import import_from_universe
    

    
def make_protein_database(logger, refmodel, refproteome):
    # make a Diamond protein database, formed by the universal protein + the reference proteins (if any).
    
    
    # log some message
    logger.debug("Creating the BiGG gene database...")
    
    
    # open the BiGG proteins fasta:
    bigg_proteins = []
    with resources.path("gempipe.assets", "bigg_proteins.faa") as asset_path:  # taken from CarveMe v1.5.2
        bigg_proteins = list(SeqIO.parse(asset_path, 'fasta'))
    
    
    # open the reference proteins (if a reference was provided):
    ref_proteins = []
    if refmodel != '-' and refproteome != '-':
        ref_proteins = list(SeqIO.parse(refproteome, 'fasta'))
            
        # read the reference model to retain just modeled gened
        refmodel = read_refmodel(refmodel)   # handle variuos formats
        if type(refmodel)==int: return 1  # an error was raised
        modeled_gids = [g.id for g in refmodel.genes]  # get the modeled genes ID
        ref_proteins = [seq_record for seq_record in ref_proteins if seq_record.id in modeled_gids]
        
        # apply the same syntax 'model.gene':
        for seq_record in ref_proteins:  
            seq_record.id = f'reference.{seq_record.id}'
            seq_record.description = ''
        
        
    # create the diamond database on the concat of the seqs:
    db_proteins = 'working/free/protein_database'
    combined_sequences = bigg_proteins + ref_proteins
    _ = SeqIO.write(combined_sequences, f'{db_proteins}.faa', 'fasta')  


    # now create the diamond database: 
    with open(f'working/logs/stdout_dmnddb.txt', 'w') as stdout, open(f'working/logs/stderr_dmnddb.txt', 'w') as stderr: 
        command = f"""diamond makedb --in {db_proteins}.faa -d {db_proteins}.dmnd"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
        
        
    return 0
            
    

def run_bigg_aligner(logger, cores):
    
    
    # some log messages
    logger.debug("Aligning to the BiGG genes database...")
    
    
    # run the command:
    with open(f'working/logs/stdout_biggalign.txt', 'w') as stdout, open(f'working/logs/stderr_biggalign.txt', 'w') as stderr: 
        command = f"""diamond blastp --threads {cores} \
            -d working/free/protein_database.dmnd \
            -q working/annotation/representatives.faa \
            -o working/free/alignment.tsv \
            --ultra-sensitive --top 10 --quiet \
            --outfmt 6 {get_blast_header()}"""
        process = subprocess.Popen(command, shell=True, stdout=stdout, stderr=stderr)
        process.wait()
        
        
        
def filter_alignment(logger, identity, coverage):
    
    
    # some log messages: 
    logger.debug("Filtering based on alignment quality...")
    logger.debug(f"Using inputted thresholds: identity {identity}%, coverage {coverage}%.") 

    
    # read the alignment
    header = get_blast_header().split(' ')
    alignment = pnd.read_csv('working/free/alignment.tsv', sep='\t', names=header)


    # filter based on alignment quality: 
    query = f'pident >= {identity} & qcovhsp >= {coverage} & scovhsp >= {coverage}'
    alignment_filtered = alignment.query(query)
    logger.debug(f"HSPs reduction: {len(alignment)} -> {len(alignment_filtered)}.")
    
    
    return alignment_filtered
    
    
    
def get_gene_combinations(r):
    # Function to dissect a GPR in its minimal genetic requisites.
    # (mainly taken from https://stackoverflow.com/a/31412649)

    
    gpr = r.gene_reaction_rule
    # Temporary replacement of special chars
    # (eg '-' would be interpreted as minus sign in eval()).
    gpr = gpr.replace('-', '___gp___MINUS___gp___')  
    gpr = gpr.replace('.', '___gp___DOT___gp___')
    if gpr == '':
        return [[]]
    

    # STEP 1. get all involved genes in this reaction.
    # EXAMPLE: reaction with id DALTA, having GPR: 
    # '(group_10845 and group_17392 and group_32377 and group_10133) or (group_10845 and group_17392 and group_10133)'
    # We can see at first sight that the minimum requirement is: 
    # (group_10845 and group_17392 and group_10133)
    involved = sorted([g.id for g in r.genes])
    # Temporary replacement of special chars
    # (eg '-' would be interpreted as minus sign in eval()).
    involved = [i.replace('-', '___gp___MINUS___gp___') for i in involved]
    involved = [i.replace('.', '___gp___DOT___gp___') for i in involved]
    
    
    # below we set a threshold to avoid computation explosion.
    # For example, if there are 34 different genes in this GPR, 
    # then the combinations to evaluate will be: 2^34 = 17,179,869,184.
    # we set max combinations =  2^15 = 32.768 
    if len(involved) > 15:
        return [involved]

    
    # STEP 2. get all combinations of involved genes together with their possible presence/absence.
    # In the above GPR, the number of different genes involved is 4.
    # This step creates a dict of {gene_A: bool, gene_B: bool, ...} for all combinations of True/False.
    # Given that bool flavours are 2 (True, False) the number of possible combinations will be
    # 2^N = 2^4 = 16, were N is the number of different genes involved (4 in this case).
    # Here a put just the first 5 combinations as example: 
    # {'group_32377': False, 'group_10845': False, 'group_10133': False, 'group_17392': False}, 
    # {'group_32377': False, 'group_10845': False, 'group_10133': False, 'group_17392': True}, 
    # {'group_32377': False, 'group_10845': False, 'group_10133': True, 'group_17392': False}, 
    # {'group_32377': False, 'group_10845': False, 'group_10133': True, 'group_17392': True}, 
    # {'group_32377': False, 'group_10845': True, 'group_10133': False, 'group_17392': False}, 
    mappings = (dict(zip(involved, values)) for values in itertools.product([False, True], repeat=len(involved)))

    
    # STEP 3. select combinations that make the whole gpr true.
    # In the example above, only 2 combination of genes satisfy the GPR: 
    # ['group_10133', 'group_10845', 'group_17392'], 
    # ['group_10133', 'group_10845', 'group_17392', 'group_32377']
    solutions = [sorted(name for name in mapping if mapping[name]) for mapping in mappings if eval(gpr, None, mapping)]

    
    # STEP 4. filter out redundant solutions.
    # In the example above, the second combination is redundant because, 
    # if it's satified, then also the first is satisfied for sure, but it's shorter. 
    # So the first combination is sufficient (minimal). 
    filtered = sorted(s1 for s1 in solutions if not any(set(s1) > set(s2) for s2 in solutions))

    
    # filtered is a list of minimum gene sets that enable the reaction
    def restore_specials(k):
        k = k.replace('___gp___MINUS___gp___', '-')
        k = k.replace('___gp___DOT___gp___', '.')
        return k
    filtered = [[restore_specials(k) for k in i] for i in filtered]   # restore original gene names
    return filtered
 
    

def get_gprm_table(logger, refmodel=None): 
    
    
    # some log messages
    logger.debug("Storing all possible gene combinations for each reaction in each model... ") 
    
    
    # read the table from assets: 
    with resources.path("gempipe.assets", "bigg_gprs.csv") as asset_path:  # taken from CarveMe v1.5.2
        # GPRM: gene to protein_complex to reaction to model
        gprm_table = pnd.read_csv(asset_path)
        
        # Some observations: 
        # A) the 'protein' column must be interpreted as 'protein complex', because it can involve several components.
        # B) Each BiGG gene can appear several times because it can be involved in several protein complexes.
        # C) Each 'protein complex' can appear several times because different reactions can require the same protein complex. 
        # D1) Each reaction can appear several times because different protein complex of the same model can enable it
        # D2) Each reaction can appear several times because different models can have the same reaction .
        
        
    # create the reference appendix for the 'gprs' table:
    if refmodel != '-':
        refmodel = read_refmodel(refmodel)   # handle variuos formats
        if type(refmodel)==int: return 1  # an error was raised
    
    
        gprm_appendix = []
        for r in refmodel.reactions: 

            # exclude exchange reactions / demand / sinks
            if len(r.metabolites) == 1:
                continue

            # create a GPR table in the style of CarveMe's 'bigg_gprs.csv'
            for comb in get_gene_combinations(r):

                if comb==[]:  # spontaneous OR reaction without GPR
                    gprm_appendix.append(
                        {'gene': f'G_refspont', 'protein': f'P_refspont', 'reaction': f'R_{r.id}', 'model': 'reference'})

                for gid in comb:
                    gprm_appendix.append(
                        {'gene': f'G_{gid}', 'protein': f'P_{"+".join(comb)}', 'reaction': f'R_{r.id}', 'model': 'reference'})

        # save appendix to file: 
        gprm_appendix = pnd.DataFrame.from_records(gprm_appendix)
        gprm_appendix.to_csv('working/free/gprm_appendix.csv')
        
        # concat the appendix
        gprm_table = pnd.concat([gprm_table, gprm_appendix], axis=0)
    
    
    # create a new column Model.Gene, like 'iCN718.ABAYE_RS01830'
    gprm_table['BiGG_gene'] = gprm_table.apply(lambda row: f"{row['model']}.{row['gene'][2:]}", axis=1) 
    
    
    return gprm_table
    
    
    
def get_gene_scores_table(logger, alignment, gprs):
    
    
    # some log messages
    logger.debug("Creating the 'gene scores' table... ") 
    
    
    def agglomerate_isoenzymes(annotation_group):
        
        # the output will be a single row, so take the first row and start to edit it. 
        res_row = annotation_group.iloc[0].copy() 
        
        # get equally good clusters for this bigg.gene
        query_genes = annotation_group['query_gene'].to_list()
        
        # when just 1 cluster: 
        if len(query_genes) == 1:
            res_row['query_gene'] = query_genes[0]
            
        else: # when many alternative clusters: 
            res_row['query_gene'] = '(' + ' or '.join(sorted(query_genes)) + ')' 
            
        # take the best performing: 
        res_row['score'] = max(annotation_group['score'].to_list())
        return res_row
    
    
    # agglomerate isoenzymes (alternative clusters for the same bigg.gene):
    annotation = alignment[['qseqid', 'sseqid', 'bitscore']]
    annotation.columns = ['query_gene', 'BiGG_gene', 'score']  # to match CarveMe names
    gene2gene = annotation \
        .sort_values(by='score', ascending=False) \
        .groupby('BiGG_gene', as_index=False) \
        .apply(agglomerate_isoenzymes)

    
    # merge with the gprs table
    gene_scores = pnd.merge(gene2gene, gprs, how='right')
    
    
    # manage spontaneus genes: 
    spontaneous_score = 0.0
    spontaneous = {'G_s0001', 'G_S0001', 'G_s_0001', 'G_S_0001', 'G_KPN_SPONT', 'G_PP_s0001'}
    spontaneous.add('G_refspont') # spontaneous reactions coming dfrom the reference model (if any)
    gene_scores.loc[gene_scores.gene.isin(spontaneous), 'score'] = spontaneous_score
    gene_scores.loc[gene_scores.gene.isin(spontaneous), 'query_gene'] = 'spontaneous'
    
    
    return gene_scores



def get_protein_scores_table(logger, gene_scores): 
    
    
    # some log messages
    logger.debug("Creating the 'protein scores' table... ")
    
    
    def merge_subunits(genes):

        n_subunits = len(genes)
        genes = genes.dropna()
        
        if len(genes) != n_subunits:
            return None

        else:
            protein = ' and '.join(sorted(genes))
            if len(genes) > 1:
                return '(' + protein + ')'
            else:
                return protein            

            
    def merge_subunit_scores(scores):
        return scores.fillna(0).mean()

    
    # WARNING ! One of the consequence of this fix, is that in the 'protein_scores' could 
    # appear protein_complexes with GPR == None and score != 0.
    # In the following step, the algorithm selects the max 'score' for the whole reaction
    # (see 'return scores.max()' in merge_protein_scores()).
    # So a protein_complex-reaction-model with GPR == None could have the max 'score'. 
    # This guarantees no changes in the original final reaction scores , while providing just 
    # the right GPR.

    
    # from gene to protein scores
    protein_scores = gene_scores.groupby(['protein', 'reaction', 'model'], as_index=False) \
        .agg({'query_gene': merge_subunits, 'score': merge_subunit_scores})
    
    
    # rename to form the GPR column 
    protein_scores.rename(columns={'query_gene': 'GPR'}, inplace=True)
    
    
    return protein_scores



def get_reaction_scores_table(logger, protein_scores):
    
    
    # some log messages
    logger.debug("Creating the 'reaction scores' table... ")
    
    
    def merge_proteins(proteins):
        
        proteins = set(proteins.dropna())
        if not proteins:
            return None
        
        gpr_str = ' or '.join(sorted(proteins))
        
        if len(proteins) > 1:
            return '(' + gpr_str + ')'
        else:
            return gpr_str

    def merge_protein_scores(scores):
        return scores.max(skipna=True)

    
    # from protein to reaction scores
    reaction_scores = protein_scores.groupby(['reaction'], as_index=False) \
        .agg({'GPR': merge_proteins, 'score': merge_protein_scores}).dropna()

    
    # Note: this table does not cotain GPR == '', because 
    # spontaneous reactions are under the 'spontaneous' gene.
    reaction_scores = reaction_scores.sort_values(by='score', ascending=False)
    reaction_scores = reaction_scores.reset_index(drop=True)
    return reaction_scores

    
    
def normalize_reaction_scores(reaction_scores):
    
    
    # take only reactions with score > 0, and get their median: 
    median_score = reaction_scores.query('score > 0')['score'].median()
    reaction_scores['normalized_score'] = (reaction_scores['score'] / median_score).apply(lambda x: round(x, 1))
    
    
    # Note: this table does not cotain GPR == '', because 
    # spontaneous reactions are under the 'spontaneous' gene.
    reaction_scores = reaction_scores.sort_values(by='normalized_score', ascending=False)
    reaction_scores = reaction_scores.reset_index(drop=True)
    return reaction_scores
    
    

def get_universe_template(logger=None, staining='neg'): 
    
    
    # some log messages
    if logger != None:  # function used also outside the pipe
        logger.debug(f"Copying the gram {staining} universe... ")
    
    
    # get a copy of the appropriate universe: 
    if staining == 'pos':  
        with resources.path("gempipe.assets", "universe_grampos.json") as asset_path:  # taken from CarveMe v1.5.2
            universe = cobra.io.load_json_model(asset_path)  
    if staining == 'neg': 
        with resources.path("gempipe.assets", "universe_gramneg.json") as asset_path:  # taken from CarveMe v1.5.2
            universe = cobra.io.load_json_model(asset_path)
            
 
    # correct compartments:
    for m in universe.metabolites:
        if   m.compartment == 'C_c':  m.compartment = 'c'
        elif m.compartment == 'C_p':  m.compartment = 'p'
        elif m.compartment == 'C_e':  m.compartment = 'e'
    universe.compartments = {'c': 'cytosol', 'p': 'periplasm', 'e': 'extracellular'}
           
        
    # some log messages
    if logger != None:  # function used also outside the pipe
        logger.debug(f"Done, {' '.join(['G:', str(len(universe.genes)), '|', 'R:', str(len(universe.reactions)), '|', 'M:', str(len(universe.metabolites))])}.")
    
    
    return universe
    
    

def perform_universe_pruning(logger, universe, reaction_scores): 
    # this process will retain the Growth reaction.
    # this process will retain also spontaneous reactions !!!
    # At this level, reactions from refmodel will be totally ignored. 
    
    
    # some log messages
    logger.debug(f"Keeping reactions with genetic support...")
    
    
    # get list of reactions with genetic support: 
    # Remove the 'R_' prefix (eg R_11M3ODO -> 11M3ODO)
    tabled_rids = list(reaction_scores['reaction'].str[2:])
    
    
    # edit the universe:
    to_remove = []
    for r in universe.reactions: 
        if r.id == "Growth": continue
        if r.id in tabled_rids: 
            gpr = reaction_scores[reaction_scores['reaction'] == "R_" + r.id]['GPR'].iloc[0]
            r.gene_reaction_rule = gpr
            r.update_genes_from_gpr()
        else: 
            to_remove.append(r)
    
    
    # remove_orphans: Remove orphaned genes and metabolites from the model as well (default False).
    universe.remove_reactions(to_remove, remove_orphans=True)
    
    
    # some log messages
    logger.debug(f"Done, {' '.join(['G:', str(len(universe.genes)), '|', 'R:', str(len(universe.reactions)), '|', 'M:', str(len(universe.metabolites))])}.")
    
    
    # substracted universe:
    universe.id = 'draft_panmodel'
    return universe
    
    

def add_exchange_reactions(logger, draft_panmodel):
    
    
    # some log messages
    logger.debug(f"Adding the exchange reactions... ")
    
    
    # Now add the exchange reactions:
    for m in draft_panmodel.metabolites: 
        
        if m.id.rsplit('_', 1)[1] == 'e': 
            
            adding = cobra.Reaction(f"EX_{m.id}")
            adding.name = f"Exchange for {m.name}"
            
            # effectively add the reaction
            draft_panmodel.add_reactions([adding])
            added = draft_panmodel.reactions.get_by_id(f"EX_{m.id}")
            
            # define the manually corrected reaction:
            r_string = f"{m.id} -->"
            added.build_reaction_from_string(r_string) 
            
            # set bounds (just for EX_reactions we specify bounds):
            added.bounds = (0, 1000)
    
    
    # some log messages:
    logger.debug(f"Done, {' '.join(['G:', str(len(draft_panmodel.genes)), '|', 'R:', str(len(draft_panmodel.reactions)), '|', 'M:', str(len(draft_panmodel.metabolites))])}.")
    
    
    return draft_panmodel
    


def eggnogg_gpr_inflator(logger, panmodel): 
    
    
    # some log messages:
    logger.debug(f"Inflating the GPRs using functional annotation...")
    
    
    # load functional annotation:
    annot = pnd.read_csv('working/annotation/pan.emapper.annotations', sep='\t', comment='#', header=None)
    annot.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')


    # create an editable copy of the model
    panmodel_update = panmodel.copy()

    
    # create a single column with all the important annotations: 
    annot = annot.set_index('query', drop=True, verify_integrity=True)
    discriminative_fields = ['Preferred_name', 'KEGG_ko', 'KEGG_Reaction', 'EC', 'KEGG_TC']
    annot['dense'] = ''
    for col in discriminative_fields:
        annot['dense'] = annot['dense'] + annot[col] 
        
    
    # divide in groups based on the dense annotation: 
    groups = annot.groupby(['dense']).groups

    
    #  get gids modeled AND annotated: 
    modeled_gids = set([g.id for g in panmodel.genes]) - set(['spontaneous'])  # all modeled genes
    annotated_gids = set(annot.index.to_list())  # annotated genes by eggnog-mapper
    modeled_annotated_gids = modeled_gids.intersection(annotated_gids)

    
    # track the parsed genes: 
    parsed_gids = set()
    
    
    # ignore the modeled genes that had not been annotated (group == '-----'): 
    group_not_annotated = ''.join(['-' for field in discriminative_fields])
    alts_not_annotated = annot.loc[groups[group_not_annotated], ]
    for gid in alts_not_annotated.index:
        if gid in modeled_gids: 
            parsed_gids.add(gid)
            
    
    # function to create the main log strings:
    def alts_to_string(alts, discriminative_fields):
        outstring = ''
        for index in alts.index:
            outstring = outstring + '\t' + index + ': ' + str([alts.loc[index, field] for field in discriminative_fields]) + '\n'
        return outstring
    
    
    # create a log subdirectory
    os.makedirs('working/free/gpr_inflator', exist_ok=True)
    
    
    # w_handler1: gids are divided by the number of alternatives.
    # w_handler2: gids are not divided (all together). 
    w_handler2 = open(f'working/free/gpr_inflator/gid_to_cluster_all.txt', 'w') 
    
    
    # create a gid to alternatives dictionary, for later use: 
    gid_to_alts = {}
    
    
    # separate the reactions based on the number of discovered isoforms starting from 1:
    admitted_alts = 1  
    while len(modeled_annotated_gids - parsed_gids) >= 1: 

        
        # Number of admitted isoforms with the exact same eggnog-mapper annotation: 
        w_handler1 = open(f'working/free/gpr_inflator/gid_to_cluster_{admitted_alts}.txt', 'w')
        

        # iterate the modeled genes:
        cnt = 0
        for g in panmodel.genes:
            gid = g.id
            if gid == 'spontaneous':
                continue

                
            # if this modeled gene has been annotated by eggnog-mapper: 
            if gid in annotated_gids:
                
                
                # get the group 
                group = annot.loc[gid, 'dense']
                
                
                # if this gene was not annotated (group == '-----'): 
                if group == ''.join(['-' for field in discriminative_fields]):
                    continue
                    
                
                # get the alternative genes: 
                alts = annot.loc[groups[group], ]

                
                # skip if different number of alternatives: 
                if len(alts) != admitted_alts:
                    continue

                
                # track the parsed gid
                cnt += 1
                parsed_gids.add(gid)
                
                
                # write the textual logs
                print(gid, '\n', alts_to_string(alts, discriminative_fields), end='', file=w_handler1)
                print(gid, '\n', alts_to_string(alts, discriminative_fields), end='', file=w_handler2)
                
                
                # populate the gid_to_alts dict: 
                if admitted_alts > 1: 
                    gid_to_alts[gid] = set(alts.index.to_list())
                  
        
        # close the handlers and increase the number of admitted alternatives
        w_handler1.close()
        admitted_alts += 1
    w_handler2.close()

    
    # iterate each reaction and edit the GPR accordingly:
    w_handler = open(f'working/free/gpr_inflator/gpr_updates.txt', 'w')
    for r in panmodel_update.reactions: 

        
        # get the gpr: 
        gpr = r.gene_reaction_rule
        if gpr == 'spontaneous': 
            continue

            
        # force each gid to be surrounded by spaces: 
        gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '

        
        # for each gid appearing in this GPR, sobstitute it with its group of alternatives: 
        involved_gids = [g.id for g in r.genes]
        for gid in involved_gids:
            if gid in gid_to_alts.keys(): 
                gpr = gpr.replace(f' {gid} ', f' editing_{gid} ')
        for gid in involved_gids:
            if gid in gid_to_alts.keys(): 
                or_group = ' or '.join(gid_to_alts[gid])
                gpr = gpr.replace(f' editing_{gid} ', f' ({or_group}) ')

                
        # remove spaces between parenthesis
        gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
        # remove spaces at the extremes: 
        gpr = gpr[1: -1]

        
        # if the GPR changed, update the model and track the logs: 
        if gpr != r.gene_reaction_rule:
            print(r.id, file=w_handler)
            print('\toriginal:', r.gene_reaction_rule, file=w_handler)
            r.gene_reaction_rule = gpr
            # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
            r.update_genes_from_gpr()
            print('\tupdated:', r.gene_reaction_rule, file=w_handler)
    w_handler.close()

    
    # some log messages:
    logger.debug(f"Done, {' '.join(['G:', str(len(panmodel_update.genes)), '|', 'R:', str(len(panmodel_update.reactions)), '|', 'M:', str(len(panmodel_update.metabolites))])}.")
    
    
    return panmodel_update



def network_rec(logger, cores, staining, identity, coverage, refmodel, refproteome):
    
    
    # create subdirs without overwriting
    os.makedirs('working/free/', exist_ok=True)
    
    
    # some log messages
    logger.info('Now starting the reference-free reconstruction...')
    
    
    # check if the needed files are already computed
    if os.path.exists('working/free/proc_acc.pickle'):
        with open('working/free/proc_acc.pickle', 'rb') as handler:
            proc_acc = pickle.load(handler) 
        if get_retained_accessions() == proc_acc:
            files_to_check = [
                'working/free/alignment.tsv',
                f'working/free/draft_panmodel_{identity}_{coverage}.json',
                'working/free/gpr_inflator/gid_to_cluster_all.txt',
                'working/free/gpr_inflator/gpr_updates.txt',
                'working/free/reaction_scores.csv',
                'working/free/reaction_scores_normalized.csv',
                'working/free/protein_scores.csv',
                'working/free/gene_scores.csv',
                'working/free/gprm_appendix.csv',
                'working/free/gprm_table.csv',
                'working/free/protein_database.faa',
                'working/free/protein_database.dmnd']
            if all([os.path.exists(file) for file in files_to_check]):
                # log some message: 
                logger.info('Found all the needed files already computed. Skipping this step.')
                # signal to skip this module:
                return 0

    
    # create the protein database:
    response = make_protein_database(logger, refmodel, refproteome)
    if response == 1: return 1
    
    
    # align representatives sequences on BiGG genes:
    run_bigg_aligner(logger, cores)
    
    
    # filter HSPs based on inputted thresholds
    alignment_filtered = filter_alignment(logger, identity, coverage)
    
    
    # get the list of all gene combinations that satisfy each reaction in each model
    gprm_table = get_gprm_table(logger, refmodel)
    if type(gprm_table) == int: return 1
    gprm_table.to_csv('working/free/gprm_table.csv')
    
    
    # get the 'gene_scores' table:
    gene_scores = get_gene_scores_table(logger, alignment_filtered, gprm_table)
    gene_scores.to_csv('working/free/gene_scores.csv')
    
    
    # get the 'protein_scores' table: 
    protein_scores = get_protein_scores_table(logger, gene_scores)
    protein_scores.to_csv('working/free/protein_scores.csv')
    
    
    # get the 'reaction_scores' table: 
    reaction_scores = get_reaction_scores_table(logger, protein_scores)
    reaction_scores.to_csv('working/free/reaction_scores.csv')
    
    
    # normalize reaction scores:
    reaction_scores_normalized = normalize_reaction_scores(reaction_scores)
    reaction_scores_normalized.to_csv('working/free/reaction_scores_normalized.csv')
    
    
    # copy the gram pos / neg universe for subtraction: 
    universe = get_universe_template(logger, staining)
    
    
    # remove reactions not present
    draft_panmodel = perform_universe_pruning(logger, universe, reaction_scores_normalized)
    
    
    # add exchange reactions
    draft_panmodel = add_exchange_reactions(logger, draft_panmodel)
    
    
    # run gene recovery via functional annotation reading
    draft_panmodel = eggnogg_gpr_inflator(logger, draft_panmodel) 
    
    
    # add the ATPM reaction for reference-free reconstructions
    if refmodel == '-' and refproteome == '-':
        universe = get_universe_template(logger, staining)   # reload to assure ATPM presence
        import_from_universe(draft_panmodel, universe, 'ATPM')        
    
    
    # save draft pan-model to disk
    cobra.io.save_json_model(draft_panmodel, f'working/free/draft_panmodel_{identity}_{coverage}.json')
    
    
    # some log messages
    logger.info(f"Reference-free reconstruction completed: {' '.join(['G:', str(len(draft_panmodel.genes)), '|', 'R:', str(len(draft_panmodel.reactions)), '|', 'M:', str(len(draft_panmodel.metabolites))])}.")
    
    
    # make traces to keep track of the accessions processed:
    # run_bigg_aligner(), first function to be called, works on annotation/representatives.faa,
    # so copy the 'proc_acc.pickle' inside annotation/
    shutil.copyfile('working/annotation/proc_acc.pickle', 'working/free/proc_acc.pickle')
    
    
    return 0